from __future__ import annotations

import os
import socket
from dataclasses import dataclass
from io import BytesIO
from typing import IO, List, Union, Optional, Dict

from cadence.frames import InitReqFrame, Frame, Arg, CallReqFrame, CallReqContinueFrame, CallResFrame, \
    CallResContinueFrame, FrameWithArgs, CallFlags, ErrorFrame
from cadence.ioutils import IOWrapper
from cadence.kvheaders import KVHeaders
from cadence.tchannel import TChannelException


class FragmentGenerator:
    def __init__(self):
        pass

    def get_args(self) -> List[bytes]:
        raise NotImplementedError()

    def get_initial_frame(self) -> FrameWithArgs:
        raise NotImplementedError()

    def get_continue_frame(self) -> FrameWithArgs:
        raise NotImplementedError()

    def build_frames(self, message_id) -> List[FrameWithArgs]:
        args: List[bytes] = self.get_args()
        frames = []
        while args:
            frame: FrameWithArgs = self.get_initial_frame() if not frames else self.get_continue_frame()
            frame.id = message_id

            while args and not frame.is_full():
                buf: bytes = args[0]
                n = len(buf)
                avail = frame.space_available() - 2  # two byte required for argument length
                if avail <= 0:
                    break

                to_write = n if avail >= n else avail
                arg = Arg(buf[0:to_write])
                frame.args.append(arg)

                buf = buf[to_write:]
                n = len(buf)

                # 1st and 2nd args that end at a frame boundary need an empty arg at the
                # start of the next frame
                if not n and not (len(args) > 1 and frame.is_frame_boundary()):
                    args.pop(0)
                else:
                    args[0] = buf

            if args:
                assert isinstance(frame, CallFlags) and isinstance(frame, FrameWithArgs)
                frame.set_more_fragments_follow(True)

            frames.append(frame)

        return frames


@dataclass
class ArgValue:
    value: bytes = bytes()
    complete: bool = False


class FragmentReader:
    def __init__(self):
        self.args = [ArgValue(), ArgValue(), ArgValue()]

    def get_incomplete_arg(self):
        for a in self.args:
            if not a.complete:
                return a
        return None

    def process_frame(self, frame: Union[FrameWithArgs]):
        self.on_load_frame(frame)
        frame_args_offset = 0
        current_arg = self.get_incomplete_arg()
        while current_arg and frame_args_offset < len(frame.args):
            frame_arg = frame.args[frame_args_offset]
            current_arg.value += frame_arg.buf
            assert isinstance(frame, (CallFlags, FrameWithArgs))
            if not frame.is_more_fragments_follow() or frame_args_offset + 1 < len(frame.args):
                current_arg.complete = True
            current_arg = self.get_incomplete_arg()
            frame_args_offset += 1

        if self.is_complete():
            self.on_args_complete(
                self.args[0].value,
                self.args[1].value,
                self.args[2].value,
            )

    def is_complete(self):
        for a in self.args:
            if not a.complete:
                return False
        return True

    def on_load_frame(self, frame: Union[FrameWithArgs]):
        raise NotImplementedError()

    def on_args_complete(self, arg1: bytes, arg2: bytes, arg3: bytes):
        raise NotImplementedError()


# noinspection PyAbstractClass
class ThriftArgScheme(FragmentGenerator, FragmentReader):
    def __init__(self):
        FragmentGenerator.__init__(self)
        FragmentReader.__init__(self)
        self.method_name = None              # arg1
        self.thrift_payload = None           # arg3
        self.application_headers = None      # arg2

    def on_args_complete(self, arg1: bytes, arg2: bytes, arg3: bytes):
        self.process_arg1(arg1)
        self.process_arg2(arg2)
        self.process_arg3(arg3)

    def process_arg1(self, b):
        self.method_name = str(b, "utf-8")

    def process_arg2(self, b):
        f: BytesIO = BytesIO(b)
        wrapper: IOWrapper = IOWrapper(f)
        h: KVHeaders = KVHeaders.read_kv_headers(wrapper, 2, "ThriftFunctionResponse")
        self.application_headers = h.d

    def process_arg3(self, b):
        self.thrift_payload = b

    def build_arg1(self) -> bytes:
        f = BytesIO()
        wrapper: IOWrapper = IOWrapper(f)
        wrapper.write_string(self.method_name)
        wrapper.flush()
        return f.getvalue()

    def build_arg2(self) -> bytes:
        f = BytesIO()
        wrapper: IOWrapper = IOWrapper(f)
        h = KVHeaders(self.application_headers, 2)
        h.write_headers(wrapper)
        return f.getvalue()

    def build_arg3(self) -> bytes:
        return self.thrift_payload

    def get_args(self) -> List[bytes]:
        args: List[bytes] = [self.build_arg1(), self.build_arg2(), self.build_arg3()]
        return args


class ThriftFunctionCall(ThriftArgScheme):

    service: Optional[str]
    method_name: str
    thrift_payload: bytes
    tchannel_headers: Optional[dict]
    application_headers: Dict[str, str]
    ttl: int

    @classmethod
    def create(cls, service: str, method_name: str, thrift_payload: bytes):
        o = cls()
        o.service = service
        o.method_name = method_name
        o.thrift_payload = thrift_payload
        o.tchannel_headers = cls.default_tchannel_headers()
        o.application_headers = cls.default_application_headers()
        # PollForActivityTask is hardcoded on the server to timeout at
        # 60 seconds, so the ttl needs to be slightly more so that it
        # does not fail.
        o.ttl = 61000
        return o

    @staticmethod
    def default_tchannel_headers():
        return {
            "as": "thrift",
            "re": "c",
            "cn": "cadence-client"
        }

    @staticmethod
    def default_application_headers():
        return {
            "user-name": os.getlogin(),
            "host-name": socket.gethostname(),
            # Copied from Java client
            "cadence-client-library-version": "2.2.0",
            "cadence-client-feature-version": "1.0.0"
        }

    def __init__(self):
        super().__init__()
        self.service = None
        self.tchannel_headers = None
        self.ttl = 0
        self.message_id = 0

    # Functions for frame reading

    def on_load_frame(self, frame: FrameWithArgs):
        if not frame.TYPE == CallReqFrame.TYPE:
            return
        # noinspection PyTypeChecker
        frame: CallReqFrame = frame
        self.message_id = frame.id
        self.service = frame.service
        self.ttl = frame.ttl
        self.tchannel_headers = frame.headers.d

    # Functions for frame generation

    def get_initial_frame(self) -> FrameWithArgs:
        frame: CallReqFrame = CallReqFrame()
        frame.ttl = self.ttl
        frame.service = self.service
        frame.headers.d.update(self.tchannel_headers)
        return frame

    def get_continue_frame(self) -> FrameWithArgs:
        frame: CallReqContinueFrame = CallReqContinueFrame()
        return frame


class ThriftFunctionResponse(ThriftArgScheme):

    thrift_payload: bytes
    method_name: str
    code: Optional[int]

    @classmethod
    def create(cls, code: int, thrift_payload):
        o = cls()
        o.code = code
        o.method_name = ""
        o.thrift_payload = thrift_payload
        o.tchannel_headers = cls.default_tchannel_headers()
        o.application_headers = cls.default_application_headers()
        return o

    @staticmethod
    def default_tchannel_headers():
        return {
            "as": "thrift"
        }

    @staticmethod
    def default_application_headers():
        return {
            "$rpc$-service": "dummy"
        }

    def __init__(self):
        ThriftArgScheme.__init__(self)
        self.message_id = None
        self.code = None
        self.tchannel_headers = None

    # Functions for frame reading

    def on_load_frame(self, frame: Union[FrameWithArgs]):
        if not frame.TYPE == CallResFrame.TYPE:
            return
        # noinspection PyTypeChecker
        frame: CallResFrame = frame
        self.message_id = frame.id
        self.code = frame.code
        self.tchannel_headers = frame.headers.d

    # Functions for frame generation

    def get_initial_frame(self) -> FrameWithArgs:
        frame: CallResFrame = CallResFrame()
        frame.code = self.code
        frame.headers.d.update(self.tchannel_headers)
        return frame

    def get_continue_frame(self) -> FrameWithArgs:
        frame: CallResContinueFrame = CallResContinueFrame()
        return frame


class TChannelConnection:

    file: IO
    wrapper: IOWrapper
    s: socket.socket

    @classmethod
    def open(cls, host: object, port: object) -> TChannelConnection:
        s: socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        return cls(s)

    def __init__(self, s: socket):
        self.s = s
        self.file = self.s.makefile("rwb")
        self.wrapper = IOWrapper(self.file)
        self.current_id = -1

        self.handshake()

    def new_id(self):
        self.current_id += 1
        return self.current_id

    def handshake(self):
        req: InitReqFrame = InitReqFrame()
        req.id = self.new_id()
        req.headers.d["host_port"] = "0.0.0.0:0"
        req.headers.d["process_name"] = "python-process"
        self.write_frame(req)

        res = self.read_frame()
        if res.TYPE != 0x02:
            raise Exception("Unexpected response from server")

    def write_frame(self, frame: Frame):
        frame.write(self.wrapper)
        self.wrapper.flush()

    def read_frame(self):
        frame = Frame.read_frame(self.wrapper)
        if isinstance(frame, ErrorFrame):
            raise TChannelException(error_frame=frame)
        return frame

    def close(self):
        self.s.close()

    def call_function(self, call: ThriftFunctionCall) -> ThriftFunctionResponse:
        frames = call.build_frames(self.new_id())
        for frame in frames:
            self.write_frame(frame)
        response = ThriftFunctionResponse()
        while not response.is_complete():
            frame = self.read_frame()
            if frame.TYPE not in (CallResFrame.TYPE, CallResContinueFrame.TYPE):
                raise Exception("Unexpected type: " + Frame.TYPE)
            assert isinstance(frame, FrameWithArgs)
            response.process_frame(frame)
        return response



from dataclasses import dataclass, field
from typing import Callable, Dict, Tuple
import inspect
import threading
import logging
import time

from cadence.conversions import camel_to_snake, snake_to_camel
from cadence.workflow import WorkflowMethod, SignalMethod
from cadence.workflowservice import WorkflowService

logger = logging.getLogger(__name__)


@dataclass
class WorkerOptions:
    pass


def _find_interface_class(impl_cls) -> type:
    hierarchy = list(inspect.getmro(impl_cls))
    hierarchy.reverse()
    hierarchy.pop(0)  # remove object
    for cls in hierarchy:
        for method_name, fn in inspect.getmembers(cls, predicate=inspect.isfunction):
            # first class with a "_workflow_method" is considered the interface
            if hasattr(fn, "_workflow_method"):
                return cls
    return impl_cls


def _find_metadata_field(cls, metadata_field, method_name):
    for c in inspect.getmro(cls):
        if not hasattr(c, method_name):
            continue
        m = getattr(c, method_name)
        if not hasattr(m, metadata_field):
            continue
        return getattr(m, metadata_field)
    return None


def _get_wm(cls: type, method_name: str) -> WorkflowMethod:
    metadata_field = "_workflow_method"
    return _find_metadata_field(cls, metadata_field, method_name)


def _get_sm(cls: type, method_name: str) -> SignalMethod:
    metadata_field = "_signal_method"
    return _find_metadata_field(cls, metadata_field, method_name)


@dataclass
class Worker:
    host: str = None
    port: int = None
    domain: str = None
    task_list: str = None
    options: WorkerOptions = None
    activities: Dict[str, Callable] = field(default_factory=dict)
    workflow_methods: Dict[str, Tuple[type, Callable]] = field(default_factory=dict)
    service: WorkflowService = None
    threads_started: int = 0
    threads_stopped: int = 0
    stop_requested: bool = False

    def register_activities_implementation(self, activities_instance: object, activities_cls_name: str = None):
        cls_name = activities_cls_name if activities_cls_name else type(activities_instance).__name__
        for method_name, fn in inspect.getmembers(activities_instance, predicate=inspect.ismethod):
            if method_name.startswith("_"):
                continue
            self.activities[f'{cls_name}::{camel_to_snake(method_name)}'] = fn
            self.activities[f'{cls_name}::{snake_to_camel(method_name)}'] = fn

    def register_workflow_implementation_type(self, impl_cls: type, workflow_cls_name: str = None):
        cls_name = workflow_cls_name if workflow_cls_name else _find_interface_class(impl_cls).__name__
        if not hasattr(impl_cls, "_signal_methods"):
            impl_cls._signal_methods = {}
        for method_name, fn in inspect.getmembers(impl_cls, predicate=inspect.isfunction):
            wm: WorkflowMethod = _get_wm(impl_cls, method_name)
            if wm:
                impl_fn = getattr(impl_cls, method_name)
                self.workflow_methods[wm._name] = (impl_cls, impl_fn)
                if "::" in wm._name:
                    _, method_name = wm._name.split("::")
                    self.workflow_methods[f'{cls_name}::{camel_to_snake(method_name)}'] = (impl_cls, impl_fn)
                    self.workflow_methods[f'{cls_name}::{snake_to_camel(method_name)}'] = (impl_cls, impl_fn)
                continue
            sm: SignalMethod = _get_sm(impl_cls, method_name)
            if sm:
                impl_fn = getattr(impl_cls, method_name)
                impl_cls._signal_methods[sm.name] = impl_fn
                if "::" in sm.name:
                    _, method_name = sm.name.split("::")
                    impl_cls._signal_methods[f'{cls_name}::{camel_to_snake(method_name)}'] = impl_fn
                    impl_cls._signal_methods[f'{cls_name}::{snake_to_camel(method_name)}'] = impl_fn
                continue


    def start(self):
        from cadence.activity_loop import activity_task_loop
        from cadence.decision_loop import DecisionTaskLoop
        self.threads_stopped = 0
        self.threads_started = 0
        self.stop_requested = False
        if self.activities:
            thread = threading.Thread(target=activity_task_loop, args=(self,))
            thread.start()
            self.threads_started += 1
        if self.workflow_methods:
            decision_task_loop = DecisionTaskLoop(worker=self)
            decision_task_loop.start()
            self.threads_started += 1

    def stop(self):
        self.stop_requested = True
        while self.threads_stopped != self.threads_started:
            time.sleep(5)

    def is_stop_requested(self):
        return self.stop_requested

    def notify_thread_stopped(self):
        self.threads_stopped += 1

    def get_workflow_method(self, workflow_type_name: str) -> Tuple[type, Callable]:
        return self.workflow_methods[workflow_type_name]





from __future__ import annotations

from typing import Tuple
from uuid import uuid4

import os
import socket

from cadence.thrift import cadence_thrift
from cadence.connection import TChannelConnection, ThriftFunctionCall
from cadence.errors import find_error
from cadence.conversions import copy_thrift_to_py, copy_py_to_thrift
from cadence.cadence_types import PollForActivityTaskResponse, StartWorkflowExecutionRequest, StartWorkflowExecutionResponse, \
    RegisterDomainRequest, PollForActivityTaskRequest, RespondActivityTaskCompletedRequest, DescribeTaskListResponse, \
    DescribeWorkflowExecutionRequest, DescribeWorkflowExecutionResponse, QueryWorkflowRequest, QueryWorkflowResponse, \
    ResetStickyTaskListResponse, ResetStickyTaskListRequest, RespondQueryTaskCompletedRequest, \
    ListClosedWorkflowExecutionsResponse, ListClosedWorkflowExecutionsRequest, ListOpenWorkflowExecutionsRequest, \
    ListOpenWorkflowExecutionsResponse, TerminateWorkflowExecutionRequest, SignalWithStartWorkflowExecutionRequest, \
    SignalWorkflowExecutionRequest, RequestCancelWorkflowExecutionRequest, RespondActivityTaskCanceledByIDRequest, \
    RespondActivityTaskCanceledRequest, RespondActivityTaskFailedByIDRequest, RespondActivityTaskFailedRequest, \
    RespondActivityTaskCompletedByIDRequest, RecordActivityTaskHeartbeatByIDRequest, \
    RecordActivityTaskHeartbeatResponse, RecordActivityTaskHeartbeatRequest, RespondDecisionTaskFailedRequest, \
    RespondDecisionTaskCompletedRequest, RespondDecisionTaskCompletedResponse, PollForDecisionTaskResponse, \
    PollForDecisionTaskRequest, GetWorkflowExecutionHistoryResponse, GetWorkflowExecutionHistoryRequest, \
    DeprecateDomainRequest, UpdateDomainRequest, UpdateDomainResponse, ListDomainsRequest, ListDomainsResponse, \
    DescribeDomainRequest, DescribeDomainResponse

TCHANNEL_SERVICE = "cadence-frontend"


class WorkflowService:

    @classmethod
    def create(cls, host: str, port: int):
        connection = TChannelConnection.open(host, port)
        return cls(connection)

    @classmethod
    def get_identity(cls):
        return "%d@%s" % (os.getpid(), socket.gethostname())

    def __init__(self, connection: TChannelConnection):
        self.connection = connection
        self.execution_start_to_close_timeout_seconds = 86400
        self.task_start_to_close_timeout_seconds = 120

    def thrift_call(self, method_name, request_argument):
        thrift_request_argument = copy_py_to_thrift(request_argument)
        fn = getattr(cadence_thrift.WorkflowService, method_name, None)
        assert fn
        request = fn.request(thrift_request_argument)
        request_payload = cadence_thrift.dumps(request)
        call = ThriftFunctionCall.create(TCHANNEL_SERVICE, "WorkflowService::" + method_name, request_payload)
        response = self.connection.call_function(call)
        start_response = cadence_thrift.loads(fn.response, response.thrift_payload)
        return start_response

    def call_return(self, method_name: str, request: object, expected_return_type: type) -> Tuple[object, object]:
        response = self.thrift_call(method_name, request)
        if not response.success:
            return None, find_error(response)
        return_value = copy_thrift_to_py(response.success)
        assert isinstance(return_value, expected_return_type)
        return return_value, None

    def call_void(self, method_name, request):
        response = self.thrift_call(method_name, request)
        error = find_error(response)
        return None, error

    def start_workflow(self, request: StartWorkflowExecutionRequest) -> Tuple[StartWorkflowExecutionResponse, object]:
        return self.call_return("StartWorkflowExecution", request, StartWorkflowExecutionResponse)

    def register_domain(self, request: RegisterDomainRequest) -> Tuple[None, object]:
        return self.call_void("RegisterDomain", request)

    def describe_domain(self, request: DescribeDomainRequest) -> Tuple[DescribeDomainResponse, object]:
        return self.call_return("DescribeDomain", request, DescribeDomainResponse)

    def list_domains(self, request: ListDomainsRequest) -> Tuple[ListDomainsResponse, object]:
        return self.call_return("ListDomains", request, ListDomainsResponse)

    def update_domain(self, request: UpdateDomainRequest) -> Tuple[UpdateDomainResponse, object]:
        return self.call_return("UpdateDomain", request, UpdateDomainResponse)

    def deprecate_domain(self, request: DeprecateDomainRequest) -> Tuple[None, object]:
        return self.call_void("DeprecateDomain", request)

    def get_workflow_execution_history(self, request: GetWorkflowExecutionHistoryRequest) -> \
            Tuple[GetWorkflowExecutionHistoryResponse, object]:
        return self.call_return("GetWorkflowExecutionHistory", request, GetWorkflowExecutionHistoryResponse)

    def poll_for_decision_task(self, request: PollForDecisionTaskRequest) -> \
            Tuple[PollForDecisionTaskResponse, object]:
        return self.call_return("PollForDecisionTask", request, PollForDecisionTaskResponse)

    def respond_decision_task_completed(self, request: RespondDecisionTaskCompletedRequest) -> \
            Tuple[RespondDecisionTaskCompletedResponse, object]:
        return self.call_return("RespondDecisionTaskCompleted", request, RespondDecisionTaskCompletedResponse)

    def respond_decision_task_failed(self, request: RespondDecisionTaskFailedRequest) -> Tuple[None, object]:
        return self.call_void("RespondDecisionTaskFailed", request)

    def poll_for_activity_task(self, request: PollForActivityTaskRequest) -> Tuple[PollForActivityTaskResponse, object]:
        return self.call_return("PollForActivityTask", request, PollForActivityTaskResponse)

    def record_activity_task_heartbeat(self, request: RecordActivityTaskHeartbeatRequest) -> \
            Tuple[RecordActivityTaskHeartbeatResponse, object]:
        return self.call_return("RecordActivityTaskHeartbeat", request, RecordActivityTaskHeartbeatResponse)

    def record_activity_task_heartbeat_by_id(self, request: RecordActivityTaskHeartbeatByIDRequest) -> \
            Tuple[RecordActivityTaskHeartbeatResponse, object]:
        return self.call_return("RecordActivityTaskHeartbeatByID", request, RecordActivityTaskHeartbeatResponse)

    def respond_activity_task_completed(self, request: RespondActivityTaskCompletedRequest) -> Tuple[None, object]:
        return self.call_void("RespondActivityTaskCompleted", request)

    def respond_activity_task_completed_by_id(self, request: RespondActivityTaskCompletedByIDRequest) -> Tuple[None, object]:
        return self.call_void("RespondActivityTaskCompletedByID", request)

    def respond_activity_task_failed(self, request: RespondActivityTaskFailedRequest) -> Tuple[None, object]:
        return self.call_void("RespondActivityTaskFailed", request)

    def respond_activity_task_failed_by_id(self, request: RespondActivityTaskFailedByIDRequest) -> Tuple[None, object]:
        return self.call_void("RespondActivityTaskFailedByID", request)

    def respond_activity_task_canceled(self, request: RespondActivityTaskCanceledRequest) -> Tuple[None, object]:
        return self.call_void("RespondActivityTaskCanceled", request)

    def respond_activity_task_canceled_by_id(self, request: RespondActivityTaskCanceledByIDRequest) -> \
        Tuple[None, object]:
        return self.call_void("RespondActivityTaskCanceledByID", request)

    def request_cancel_workflow_execution(self, request: RequestCancelWorkflowExecutionRequest) -> \
            Tuple[None, object]:
        return self.call_void("RequestCancelWorkflowExecution", request)

    def signal_workflow_execution(self, request: SignalWorkflowExecutionRequest) -> Tuple[None, object]:
        return self.call_void("SignalWorkflowExecution", request)

    def signal_with_start_workflow_execution(self, request: SignalWithStartWorkflowExecutionRequest) -> \
            Tuple[StartWorkflowExecutionResponse, object]:
        return self.call_return("SignalWithStartWorkflowExecution", request, StartWorkflowExecutionResponse)

    def terminate_workflow_execution(self, request: TerminateWorkflowExecutionRequest) -> Tuple[None, object]:
        return self.call_void("TerminateWorkflowExecution", request)

    def list_open_workflow_executions(self, request: ListOpenWorkflowExecutionsRequest) -> \
            Tuple[ListOpenWorkflowExecutionsResponse, object]:
        return self.call_return("ListOpenWorkflowExecutions", request, ListOpenWorkflowExecutionsResponse)

    def list_closed_workflow_executions(self, request: ListClosedWorkflowExecutionsRequest) -> \
            Tuple[ListClosedWorkflowExecutionsResponse, object]:
        return self.call_return("ListClosedWorkflowExecutions", request, ListClosedWorkflowExecutionsResponse)

    def respond_query_task_completed(self, request: RespondQueryTaskCompletedRequest) -> Tuple[None, object]:
        return self.call_void("RespondQueryTaskCompleted", request)

    def reset_sticky_task_list(self, request: ResetStickyTaskListRequest) -> Tuple[ResetStickyTaskListResponse, object]:
        return self.call_return("ResetStickyTaskList", request, ResetStickyTaskListResponse)

    def query_workflow(self, request: QueryWorkflowRequest) -> Tuple[QueryWorkflowResponse, object]:
        return self.call_return("QueryWorkflow", request, QueryWorkflowResponse)

    def describe_workflow_execution(self, request: DescribeWorkflowExecutionRequest) -> Tuple[DescribeWorkflowExecutionResponse, object]:
        return self.call_return("DescribeWorkflowExecution", request, DescribeWorkflowExecutionResponse)

    def describe_task_list(self, request) -> Tuple[DescribeTaskListResponse, object]:
        return self.call_return("DescribeTaskList", request, DescribeTaskListResponse)

    def close(self):
        self.connection.close()



from __future__ import annotations

import asyncio
import contextvars
import datetime
import json
import uuid
import random
import logging
import threading
from asyncio.base_futures import CancelledError
from asyncio.events import AbstractEventLoop
from asyncio.futures import Future
from asyncio.tasks import Task
from collections import OrderedDict
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Dict, Optional, Any, Callable

from more_itertools import peekable

from cadence.activity_method import ExecuteActivityParameters
from cadence.cadence_types import PollForDecisionTaskRequest, TaskList, PollForDecisionTaskResponse, \
    RespondDecisionTaskCompletedRequest, \
    CompleteWorkflowExecutionDecisionAttributes, Decision, DecisionType, RespondDecisionTaskCompletedResponse, \
    HistoryEvent, EventType, WorkflowType, ScheduleActivityTaskDecisionAttributes, \
    CancelWorkflowExecutionDecisionAttributes, StartTimerDecisionAttributes, TimerFiredEventAttributes
from cadence.conversions import json_to_args
from cadence.decisions import DecisionId, DecisionTarget
from cadence.exceptions import WorkflowTypeNotFound, NonDeterministicWorkflowException, ActivityTaskFailedException, \
    ActivityTaskTimeoutException, SignalNotFound
from cadence.state_machines import ActivityDecisionStateMachine, DecisionStateMachine, CompleteWorkflowStateMachine, \
    TimerDecisionStateMachine
from cadence.tchannel import TChannelException
from cadence.worker import Worker
from cadence.workflowservice import WorkflowService

logger = logging.getLogger(__name__)


def is_decision_event(event: HistoryEvent) -> bool:
    decision_event_types = (EventType.ActivityTaskScheduled,
                            EventType.StartChildWorkflowExecutionInitiated,
                            EventType.TimerStarted,
                            EventType.WorkflowExecutionCompleted,
                            EventType.WorkflowExecutionFailed,
                            EventType.WorkflowExecutionCanceled,
                            EventType.WorkflowExecutionContinuedAsNew,
                            EventType.ActivityTaskCancelRequested,
                            EventType.RequestCancelActivityTaskFailed,
                            EventType.TimerCanceled,
                            EventType.CancelTimerFailed,
                            EventType.RequestCancelExternalWorkflowExecutionInitiated,
                            EventType.MarkerRecorded,
                            EventType.SignalExternalWorkflowExecutionInitiated)
    return event.event_type in decision_event_types


def nano_to_milli(nano):
    return nano/(1000 * 1000)


class HistoryHelper:

    def __init__(self, events: List[HistoryEvent]):
        self.events = peekable(events)

    def has_next(self) -> bool:
        try:
            self.events.peek()
            return True
        except StopIteration:
            return False

    def next(self) -> Optional[DecisionEvents]:
        events = self.events
        if not self.has_next():
            return None
        decision_events: List[HistoryEvent] = []
        new_events: List[HistoryEvent] = []
        replay = True
        next_decision_event_id = -1
        # noinspection PyUnusedLocal
        event: HistoryEvent
        for event in events:
            event_type = event.event_type
            if event_type == EventType.DecisionTaskStarted or not self.has_next():
                replay_current_time_milliseconds = nano_to_milli(event.timestamp)
                if not self.has_next():
                    replay = False
                    next_decision_event_id = event.event_id + 2
                    break
                peeked: HistoryEvent = events.peek()
                peeked_type = peeked.event_type
                if peeked_type == EventType.DecisionTaskTimedOut or peeked_type == EventType.DecisionTaskFailed:
                    continue
                elif peeked_type == EventType.DecisionTaskCompleted:
                    next(events)
                    next_decision_event_id = peeked.event_id + 1
                    break
                else:
                    raise Exception(
                        "Unexpected event after DecisionTaskStarted: {}".format(peeked))
            new_events.append(event)
        while self.has_next():
            if not is_decision_event(events.peek()):
                break
            decision_events.append(next(events))
        result = DecisionEvents(new_events, decision_events, replay,
                                replay_current_time_milliseconds, next_decision_event_id)
        logger.debug("HistoryHelper next=%s", result)
        return result


@dataclass
class DecisionEvents:
    events: List[HistoryEvent]
    decision_events: List[HistoryEvent]
    replay: bool
    replay_current_time_milliseconds: int
    next_decision_event_id: int


class Status(Enum):
    CREATED = 1
    RUNNING = 2
    DONE = 3


current_task = contextvars.ContextVar("current_task")


@dataclass
class ITask:
    decider: ReplayDecider = None
    task: Task = None
    status: Status = Status.CREATED
    awaited: Future = None

    def is_done(self):
        return self.status == Status.DONE

    def destroy(self):
        if self.status == Status.RUNNING:
            self.status = Status.DONE
            self.task.cancel()

    def start(self):
        pass

    async def await_till(self, c: Callable, timeout_seconds: int = 0) -> bool:
        timer_cancellation_handler: TimerCancellationHandler = None
        timer_fired = False

        def timer_callback(ex: Exception):
            nonlocal timer_fired
            if not ex:
                timer_fired = True

        if timeout_seconds:
            timer_cancellation_handler = self.decider.decision_context.create_timer(delay_seconds=timeout_seconds, callback=timer_callback)

        while not c() and not timer_fired:
            self.awaited = self.decider.event_loop.create_future()
            await self.awaited
            assert self.awaited.done()

        self.awaited = None

        if timer_fired:
            return False

        if timer_cancellation_handler:
            timer_cancellation_handler.accept(None)

        return True

    def unblock(self):
        if self.awaited:
            self.awaited.set_result(None)

    @staticmethod
    def current() -> ITask:
        return current_task.get()


@dataclass
class WorkflowMethodTask(ITask):
    task_id: str = None
    workflow_input: List = None
    worker: Worker = None
    workflow_type: WorkflowType = None
    workflow_instance: object = None
    ret_value: object = None
    exception_thrown: BaseException = None

    def __post_init__(self):
        logger.debug(f"[task-{self.task_id}] Created")
        self.task = asyncio.get_event_loop().create_task(self.init_workflow_instance())

    async def init_workflow_instance(self):
        current_task.set(self)
        cls, _ = self.worker.get_workflow_method(self.workflow_type.name)
        try:
            self.workflow_instance = cls()
            self.task = asyncio.get_event_loop().create_task(self.workflow_main())
        except Exception as ex:
            logger.error(
                f"Initialization of Workflow {self.workflow_type.name}({str(self.workflow_input)[1:-1]}) failed", exc_info=1)
            self.exception_thrown = ex
            self.status = Status.DONE

    async def workflow_main(self):
        logger.debug(f"[task-{self.task_id}] Running")

        if self.is_done():
            return

        current_task.set(self)

        if self.workflow_type.name not in self.worker.workflow_methods:
            self.status = Status.DONE
            self.exception_thrown = WorkflowTypeNotFound(self.workflow_type.name)
            logger.error(f"Workflow type not found: {self.workflow_type.name}")
            return

        cls, workflow_proc = self.worker.workflow_methods[self.workflow_type.name]
        self.status = Status.RUNNING
        try:
            logger.info(f"Invoking workflow {self.workflow_type.name}({str(self.workflow_input)[1:-1]})")
            self.ret_value = await workflow_proc(self.workflow_instance, *self.workflow_input)
            logger.info(
                f"Workflow {self.workflow_type.name}({str(self.workflow_input)[1:-1]}) returned {self.ret_value}")
            self.decider.complete_workflow_execution(self.ret_value)
        except CancelledError:
            logger.debug("Coroutine cancelled (expected)")
        except Exception as ex:
            logger.error(
                f"Workflow {self.workflow_type.name}({str(self.workflow_input)[1:-1]}) failed", exc_info=1)
            self.exception_thrown = ex
        finally:
            self.status = Status.DONE

    def get_workflow_instance(self):
        return self.workflow_instance


@dataclass
class SignalMethodTask(ITask):
    task_id: str = None
    workflow_instance: object = None
    signal_name: str = None
    signal_input: List = None
    exception_thrown: BaseException = None
    ret_value: object = None

    def start(self):
        logger.debug(f"[signal-task-{self.task_id}-{self.signal_name}] Created")
        self.task = asyncio.get_event_loop().create_task(self.signal_main())

    async def signal_main(self):
        logger.debug(f"[signal-task-{self.task_id}-{self.signal_name}] Running")
        current_task.set(self)

        if not self.signal_name in self.workflow_instance._signal_methods:
            self.status = Status.DONE
            self.exception_thrown = SignalNotFound(self.signal_name)
            logger.error(f"Signal not found: {self.signal_name}")
            return

        signal_proc = self.workflow_instance._signal_methods[self.signal_name]
        self.status = Status.RUNNING

        try:
            logger.info(f"Invoking signal {self.signal_name}({str(self.signal_input)[1:-1]})")
            self.ret_value = await signal_proc(self.workflow_instance, *self.signal_input)
            logger.info(
                f"Signal {self.signal_name}({str(self.signal_input)[1:-1]}) returned {self.ret_value}")
            self.decider.complete_signal_execution(self)
        except CancelledError:
            logger.debug("Coroutine cancelled (expected)")
        except Exception as ex:
            logger.error(
                f"Signal {self.signal_name}({str(self.signal_input)[1:-1]}) failed", exc_info=1)
            self.exception_thrown = ex
        finally:
            self.status = Status.DONE


@dataclass
class EventLoopWrapper:
    event_loop: AbstractEventLoop = None

    def __post_init__(self):
        self.event_loop = asyncio.get_event_loop()

    def run_event_loop_once(self):
        self.event_loop.call_soon(self.event_loop.stop)
        self.event_loop.run_forever()

    def create_future(self) -> Future[Any]:
        return self.event_loop.create_future()


@dataclass
class DecisionContext:
    decider: ReplayDecider
    scheduled_activities: Dict[int, Future[bytes]] = field(default_factory=dict)
    workflow_clock: ClockDecisionContext = None
    current_run_id: str = None

    def __post_init__(self):
        if not self.workflow_clock:
            self.workflow_clock = ClockDecisionContext(self.decider)

    async def schedule_activity_task(self, parameters: ExecuteActivityParameters):
        attr = ScheduleActivityTaskDecisionAttributes()
        attr.activity_type = parameters.activity_type
        attr.input = parameters.input
        if parameters.heartbeat_timeout_seconds > 0:
            attr.heartbeat_timeout_seconds = parameters.heartbeat_timeout_seconds
        attr.schedule_to_close_timeout_seconds = parameters.schedule_to_close_timeout_seconds
        attr.schedule_to_start_timeout_seconds = parameters.schedule_to_start_timeout_seconds
        attr.start_to_close_timeout_seconds = parameters.start_to_close_timeout_seconds
        attr.activity_id = parameters.activity_id
        if not attr.activity_id:
            attr.activity_id = self.decider.get_and_increment_next_id()
        attr.task_list = TaskList()
        attr.task_list.name = parameters.task_list

        # PORT: RetryParameters retryParameters = parameters.getRetryParameters();
        # PORT: if (retryParameters != null) {
        # PORT:    attributes.setRetryPolicy(retryParameters.toRetryPolicy());
        # PORT: }

        scheduled_event_id = self.decider.schedule_activity_task(schedule=attr)
        future = self.decider.event_loop.create_future()
        self.scheduled_activities[scheduled_event_id] = future
        await future
        assert future.done()
        exception = future.exception()
        if exception:
            raise exception
        raw_bytes = future.result()
        return json.loads(str(raw_bytes, "utf-8"))

    async def schedule_timer(self, seconds: int):
        future = self.decider.event_loop.create_future()

        def callback(ex: Exception):
            nonlocal future
            if ex:
                future.set_exception(ex)
            else:
                future.set_result("time-fired")

        self.decider.decision_context.create_timer(delay_seconds=seconds, callback=callback)
        await future
        assert future.done()
        exception = future.exception()
        if exception:
            raise exception
        return

    def handle_activity_task_completed(self, event: HistoryEvent):
        attr = event.activity_task_completed_event_attributes
        if self.decider.handle_activity_task_closed(attr.scheduled_event_id):
            future = self.scheduled_activities.get(attr.scheduled_event_id)
            if future:
                self.scheduled_activities.pop(attr.scheduled_event_id)
                future.set_result(attr.result)
            else:
                raise NonDeterministicWorkflowException(
                    f"Trying to complete activity event {attr.scheduled_event_id} that is not in scheduled_activities")

    def handle_activity_task_failed(self, event: HistoryEvent):
        attr = event.activity_task_failed_event_attributes
        if self.decider.handle_activity_task_closed(attr.scheduled_event_id):
            future = self.scheduled_activities.get(attr.scheduled_event_id)
            if future:
                self.scheduled_activities.pop(attr.scheduled_event_id)
                ex = ActivityTaskFailedException(attr.reason, attr.details)
                future.set_exception(ex)
            else:
                raise NonDeterministicWorkflowException(
                    f"Trying to complete activity event {attr.scheduled_event_id} that is not in scheduled_activities")

    def handle_activity_task_timed_out(self, event: HistoryEvent):
        attr = event.activity_task_timed_out_event_attributes
        if self.decider.handle_activity_task_closed(attr.scheduled_event_id):
            future = self.scheduled_activities.get(attr.scheduled_event_id)
            if future:
                self.scheduled_activities.pop(attr.scheduled_event_id)
                ex = ActivityTaskTimeoutException(event.event_id, attr.timeout_type, attr.details)
                future.set_exception(ex)
            else:
                raise NonDeterministicWorkflowException(
                    f"Trying to complete activity event {attr.scheduled_event_id} that is not in scheduled_activities")

    def create_timer(self, delay_seconds: int, callback: Callable):
        return self.workflow_clock.create_timer(delay_seconds, callback)

    def set_replay_current_time_milliseconds(self, replay_current_time_milliseconds: int):
        if replay_current_time_milliseconds < self.workflow_clock.current_time_millis():
            raise Exception("workflow clock moved back")
        self.workflow_clock.set_replay_current_time_milliseconds(replay_current_time_milliseconds)

    def current_time_millis(self):
        return self.workflow_clock.current_time_millis()

    def set_replaying(self, replaying: bool):
        self.workflow_clock.set_replaying(replaying)

    def is_replaying(self):
        return self.workflow_clock.is_replaying()

    def handle_timer_fired(self, attributes: TimerFiredEventAttributes):
        self.workflow_clock.handle_timer_fired(attributes)

    def handle_timer_canceled(self, event: HistoryEvent):
        self.workflow_clock.handle_timer_canceled(event)

    def set_current_run_id(self, run_id: str):
        self.current_run_id = run_id

    def random_uuid(self) -> uuid.UUID:
        return uuid.uuid3(uuid.UUID(self.current_run_id), str(self.decider.get_and_increment_next_id()))
    
    def new_random(self) -> random.Random:
        random_uuid = self.random_uuid()
        lsb = random_uuid.bytes[:8]
        generator = random.Random()
        generator.seed(lsb, version=2)
        return generator


@dataclass
class ReplayDecider:
    execution_id: str
    workflow_type: WorkflowType
    worker: Worker
    workflow_task: WorkflowMethodTask = None
    tasks: List[ITask] = field(default_factory=list)
    event_loop: EventLoopWrapper = field(default_factory=EventLoopWrapper)
    completed: bool = False

    next_decision_event_id: int = 0
    id_counter: int = 0
    decision_events: DecisionEvents = None
    decisions: OrderedDict[DecisionId, DecisionStateMachine] = field(default_factory=OrderedDict)
    decision_context: DecisionContext = None

    activity_id_to_scheduled_event_id: Dict[str, int] = field(default_factory=dict)

    def __post_init__(self):
        self.decision_context = DecisionContext(decider=self)

    def decide(self, events: List[HistoryEvent]):
        helper = HistoryHelper(events)
        while helper.has_next():
            decision_events = helper.next()
            self.process_decision_events(decision_events)
        return self.get_decisions()

    def process_decision_events(self, decision_events: DecisionEvents):
        self.decision_context.set_replaying(decision_events.replay)
        self.decision_context.set_replay_current_time_milliseconds(decision_events.replay_current_time_milliseconds)

        self.handle_decision_task_started(decision_events)
        for event in decision_events.events:
            self.process_event(event)
        if self.completed:
            return
        self.unblock_all()
        self.event_loop.run_event_loop_once()
        if decision_events.replay:
            self.notify_decision_sent()
        for event in decision_events.decision_events:
            self.process_event(event)

    def unblock_all(self):
        for t in self.tasks:
            t.unblock()

    def process_event(self, event: HistoryEvent):
        event_handler = event_handlers.get(event.event_type)
        if not event_handler:
            raise Exception(f"No event handler for event type {event.event_type.name}")
        event_handler(self, event)

    def handle_workflow_execution_started(self, event: HistoryEvent):
        start_event_attributes = event.workflow_execution_started_event_attributes
        self.decision_context.set_current_run_id(start_event_attributes.original_execution_run_id)
        if start_event_attributes.input is None:
            workflow_input = []
        else:
            workflow_input = json_to_args(start_event_attributes.input)
        self.workflow_task = WorkflowMethodTask(task_id=self.execution_id, workflow_input=workflow_input,
                                                worker=self.worker, workflow_type=self.workflow_type, decider=self)
        self.event_loop.run_event_loop_once()
        assert self.workflow_task.workflow_instance
        self.tasks.append(self.workflow_task)

    def handle_workflow_execution_cancel_requested(self, event: HistoryEvent):
        self.cancel_workflow_execution()

    def notify_decision_sent(self):
        for state_machine in self.decisions.values():
            if state_machine.get_decision():
                state_machine.handle_decision_task_started_event()

    def handle_decision_task_started(self, decision_events: DecisionEvents):
        self.decision_events = decision_events
        self.next_decision_event_id = decision_events.next_decision_event_id

    def complete_workflow_execution(self, ret_value):
        # PORT: addAllMissingVersionMarker(false, Optional.empty());
        decision = Decision()
        attr = CompleteWorkflowExecutionDecisionAttributes()
        attr.result = json.dumps(ret_value)
        decision.complete_workflow_execution_decision_attributes = attr
        decision.decision_type = DecisionType.CompleteWorkflowExecution
        decision_id = DecisionId(DecisionTarget.SELF, 0)
        self.add_decision(decision_id, CompleteWorkflowStateMachine(decision_id, decision))
        self.completed = True

    def cancel_workflow_execution(self):
        logger.info("Canceling workflow: %s", self.execution_id)
        decision = Decision()
        attr = CancelWorkflowExecutionDecisionAttributes()
        attr.details = None
        decision.cancel_workflow_execution_decision_attributes = attr
        decision.decision_type = DecisionType.CancelWorkflowExecution
        decision_id = DecisionId(DecisionTarget.SELF, 0)
        self.add_decision(decision_id, CompleteWorkflowStateMachine(decision_id, decision))
        self.completed = True

    def schedule_activity_task(self, schedule: ScheduleActivityTaskDecisionAttributes) -> int:
        # PORT: addAllMissingVersionMarker(false, Optional.empty());
        next_decision_event_id = self.next_decision_event_id
        decision_id = DecisionId(DecisionTarget.ACTIVITY, next_decision_event_id)
        self.activity_id_to_scheduled_event_id[schedule.activity_id] = next_decision_event_id
        self.add_decision(decision_id, ActivityDecisionStateMachine(decision_id, schedule_attributes=schedule))
        return next_decision_event_id

    def complete_signal_execution(self, task: SignalMethodTask):
        task.destroy()
        self.tasks.remove(task)

    def handle_activity_task_closed(self, scheduled_event_id: int) -> bool:
        decision: DecisionStateMachine = self.get_decision(DecisionId(DecisionTarget.ACTIVITY, scheduled_event_id))
        assert decision
        decision.handle_completion_event()
        return decision.is_done()

    def handle_activity_task_scheduled(self, event: HistoryEvent):
        decision = self.get_decision(DecisionId(DecisionTarget.ACTIVITY, event.event_id))
        decision.handle_initiated_event(event)

    def handle_activity_task_started(self, event: HistoryEvent):
        attr = event.activity_task_started_event_attributes
        decision = self.get_decision(DecisionId(DecisionTarget.ACTIVITY, attr.scheduled_event_id))
        decision.handle_started_event(event)

    def handle_activity_task_completed(self, event: HistoryEvent):
        self.decision_context.handle_activity_task_completed(event)

    def handle_activity_task_failed(self, event: HistoryEvent):
        self.decision_context.handle_activity_task_failed(event)

    def handle_activity_task_timed_out(self, event: HistoryEvent):
        self.decision_context.handle_activity_task_timed_out(event)

    def handle_workflow_execution_signaled(self, event: HistoryEvent):
        signaled_event_attributes = event.workflow_execution_signaled_event_attributes
        signal_input = signaled_event_attributes.input
        if not signal_input:
            signal_input = []
        else:
            signal_input = json_to_args(signal_input)

        task = SignalMethodTask(task_id=self.execution_id,
                                workflow_instance=self.workflow_task.workflow_instance,
                                signal_name=signaled_event_attributes.signal_name,
                                signal_input=signal_input,
                                decider=self)
        self.tasks.append(task)
        task.start()

    def add_decision(self, decision_id: DecisionId, decision: DecisionStateMachine):
        self.decisions[decision_id] = decision
        self.next_decision_event_id += 1

    def get_and_increment_next_id(self) -> str:
        ret_value = str(self.id_counter)
        self.id_counter += 1
        return ret_value

    def get_decision(self, decision_id: DecisionId) -> DecisionStateMachine:
        result: DecisionStateMachine = self.decisions.get(decision_id)
        if not result:
            raise NonDeterministicWorkflowException(f"Unknown {decision_id}.")
        return result

    def get_decisions(self) -> List[Decision]:
        decisions = []
        for state_machine in self.decisions.values():
            d = state_machine.get_decision()
            if d:
                decisions.append(d)

        # PORT: // Include FORCE_IMMEDIATE_DECISION timer only if there are more then 100 events
        # PORT: int size = result.size();
        # PORT: if (size > MAXIMUM_DECISIONS_PER_COMPLETION &&
        # PORT:         !isCompletionEvent(result.get(MAXIMUM_DECISIONS_PER_COMPLETION - 2))) {
        # PORT:     result = result.subList(0, MAXIMUM_DECISIONS_PER_COMPLETION - 1);
        # PORT:     StartTimerDecisionAttributes attributes = new StartTimerDecisionAttributes();
        # PORT:     attributes.setStartToFireTimeoutSeconds(0);
        # PORT:     attributes.setTimerId(FORCE_IMMEDIATE_DECISION_TIMER);
        # PORT:     Decision d = new Decision();
        # PORT:     d.setStartTimerDecisionAttributes(attributes);
        # PORT:     d.setDecisionType(DecisionType.StartTimer);
        # PORT:     result.add(d);
        # PORT: }

        return decisions

    def destroy(self):
        if self.workflow_task:
            self.workflow_task.destroy()

    def start_timer(self, request: StartTimerDecisionAttributes):
        start_event_id = self.next_decision_event_id
        decision_id = DecisionId(DecisionTarget.TIMER, start_event_id)
        self.add_decision(decision_id, TimerDecisionStateMachine(decision_id, start_timer_attributes=request))
        return start_event_id

    def cancel_timer(self, start_event_id: int, immediate_cancellation_callback: Callable):
        decision: DecisionStateMachine = self.get_decision(DecisionId(DecisionTarget.TIMER, start_event_id))
        if decision.is_done():
            return True
        if decision.cancel(immediate_cancellation_callback):
            self.next_decision_event_id += 1
        return decision.is_done()

    def handle_timer_closed(self, attributes: TimerFiredEventAttributes) -> bool:
        decision = self.get_decision(DecisionId(DecisionTarget.TIMER, attributes.started_event_id))
        decision.handle_completion_event()
        return decision.is_done()

    def handle_timer_canceled(self, event: HistoryEvent) -> bool:
        attributes = event.timer_canceled_event_attributes
        decision = self.get_decision(DecisionId(DecisionTarget.TIMER, attributes.started_event_id))
        decision.handle_cancellation_event()
        return decision.is_done()

    def handle_cancel_timer_failed(self, event: HistoryEvent) -> bool:
        started_event_id = event.event_id
        decision = self.get_decision(DecisionId(DecisionTarget.TIMER, started_event_id))
        decision.handle_cancellation_failure_event(event)
        return decision.is_done()

    def handle_timer_started(self, event: HistoryEvent):
        decision = self.get_decision(DecisionId(DecisionTarget.TIMER, event.event_id))
        decision.handle_initiated_event(event)

    def handle_timer_fired(self, event: HistoryEvent):
        attributes = event.timer_fired_event_attributes
        self.decision_context.handle_timer_fired(attributes)


# noinspection PyUnusedLocal
def noop(*args):
    pass


def on_timer_canceled(self: ReplayDecider, event: HistoryEvent):
    self.decision_context.handle_timer_canceled(event)


event_handlers = {
    EventType.WorkflowExecutionStarted: ReplayDecider.handle_workflow_execution_started,
    EventType.WorkflowExecutionCancelRequested: ReplayDecider.handle_workflow_execution_cancel_requested,
    EventType.DecisionTaskScheduled: noop,
    EventType.DecisionTaskStarted: noop,  # Filtered by HistoryHelper
    EventType.DecisionTaskTimedOut: noop,  # TODO: check
    EventType.ActivityTaskScheduled: ReplayDecider.handle_activity_task_scheduled,
    EventType.ActivityTaskStarted: ReplayDecider.handle_activity_task_started,
    EventType.ActivityTaskCompleted: ReplayDecider.handle_activity_task_completed,
    EventType.ActivityTaskFailed: ReplayDecider.handle_activity_task_failed,
    EventType.ActivityTaskTimedOut: ReplayDecider.handle_activity_task_timed_out,
    EventType.WorkflowExecutionSignaled: ReplayDecider.handle_workflow_execution_signaled,
    EventType.TimerFired: ReplayDecider.handle_timer_fired,
    EventType.TimerStarted: ReplayDecider.handle_timer_started,
    EventType.TimerCanceled: on_timer_canceled,
    EventType.CancelTimerFailed: ReplayDecider.handle_cancel_timer_failed
}


@dataclass
class DecisionTaskLoop:
    worker: Worker
    service: WorkflowService = None
    deciders: Dict[str, ReplayDecider] = field(default_factory=dict)

    def __post_init__(self):
        pass

    def start(self):
        thread = threading.Thread(target=self.run)
        thread.start()

    def run(self):
        try:
            logger.info(f"Decision task worker started: {WorkflowService.get_identity()}")
            event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(event_loop)
            self.service = WorkflowService.create(self.worker.host, self.worker.port)
            while True:
                if self.worker.is_stop_requested():
                    return
                decision_task: PollForDecisionTaskResponse = self.poll()
                if not decision_task:
                    continue
                decisions = self.process_task(decision_task)
                self.respond_decisions(decision_task.task_token, decisions)
        finally:
            # noinspection PyPep8,PyBroadException
            try:
                self.service.close()
            except:
                logger.warning("service.close() failed", exc_info=1)
            self.worker.notify_thread_stopped()

    def poll(self) -> Optional[PollForDecisionTaskResponse]:
        try:
            polling_start = datetime.datetime.now()
            poll_decision_request = PollForDecisionTaskRequest()
            poll_decision_request.identity = WorkflowService.get_identity()
            poll_decision_request.task_list = TaskList()
            poll_decision_request.task_list.name = self.worker.task_list
            poll_decision_request.domain = self.worker.domain
            # noinspection PyUnusedLocal
            task: PollForDecisionTaskResponse
            task, err = self.service.poll_for_decision_task(poll_decision_request)
            polling_end = datetime.datetime.now()
            logger.debug("PollForDecisionTask: %dms", (polling_end - polling_start).total_seconds() * 1000)
        except TChannelException as ex:
            logger.error("PollForDecisionTask error: %s", ex)
            return None
        if err:
            logger.error("PollForDecisionTask failed: %s", err)
            return None
        if not task.task_token:
            logger.debug("PollForActivityTask has no task token (expected): %s", task)
            return None
        return task

    def process_task(self, decision_task: PollForDecisionTaskResponse) -> List[Decision]:
        execution_id = str(decision_task.workflow_execution)
        decider = ReplayDecider(execution_id, decision_task.workflow_type, self.worker)
        decisions: List[Decision] = decider.decide(decision_task.history.events)
        decider.destroy()
        return decisions

    def respond_decisions(self, task_token: bytes, decisions: List[Decision]):
        service = self.service
        request = RespondDecisionTaskCompletedRequest()
        request.task_token = task_token
        request.decisions.extend(decisions)
        request.identity = WorkflowService.get_identity()
        # noinspection PyUnusedLocal
        response: RespondDecisionTaskCompletedResponse
        response, err = service.respond_decision_task_completed(request)
        if err:
            logger.error("Error invoking RespondDecisionTaskCompleted: %s", err)
        else:
            logger.debug("RespondDecisionTaskCompleted: %s", response)


from cadence.clock_decision_context import ClockDecisionContext, TimerCancellationHandler



from cadence.cadence_types import TimeoutType


class IllegalStateException(BaseException):
    pass


class IllegalArgumentException(BaseException):
    pass


class WorkflowTypeNotFound(Exception):
    pass


class NonDeterministicWorkflowException(BaseException):
    pass


class ActivityTaskFailedException(Exception):

    def __init__(self, reason: str, details: bytes) -> None:
        super().__init__(reason)
        self.reason = reason
        self.details = details


class ActivityTaskTimeoutException(Exception):

    def __init__(self, event_id: int, timeout_type: TimeoutType, details: bytes, *args: object) -> None:
        super().__init__(*args)
        self.details = details
        self.timeout_type = timeout_type
        self.event_id = event_id


class SignalNotFound(Exception):
    pass


class CancellationException(Exception):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cause = None

    def init_cause(self, cause):
        self.cause = cause



from cadence.frames import ErrorFrame

ERROR_INVALID = 0x00
ERROR_TIMEOUT = 0x01
ERROR_CANCELLED = 0x02
ERROR_BUSY = 0x03
ERROR_DECLINED = 0x04
ERROR_UNEXPECTED_ERROR = 0x05
ERROR_BAD_REQUEST = 0x06
ERROR_NETWORK_ERROR = 0x07
ERROR_UNHEALTHY = 0x08
ERROR_FATAL = 0x0FF


class TChannelException(Exception):

    def __init__(self, error_frame: ErrorFrame = None):
        self.error_frame = error_frame

    def __str__(self):
        return self.error_frame.message



import json
import typing
import inspect
import re
from enum import Enum, IntEnum
import cadence.cadence_types
from cadence.thrift import cadence_thrift

PRIMITIVES = [int, str, bytes, float, bool]


def camel_to_snake(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def copy_thrift_to_py(thrift_object, field_type=None):
    if thrift_object is None:
        obj = None
    elif field_type and field_type in PRIMITIVES:
        obj = thrift_object
    elif field_type and inspect.isclass(field_type) and issubclass(field_type, Enum):
        obj = field_type.value_for(thrift_object)
    elif type(thrift_object) == list:
        assert field_type
        obj = []
        for item in thrift_object:
            obj.append(copy_thrift_to_py(item, field_type=field_type.__args__[0]))
    elif type(thrift_object) == dict:
        assert field_type
        obj = {}
        assert isinstance(thrift_object, dict)
        for key, value in thrift_object.items():
            obj[key] = copy_thrift_to_py(value, field_type=field_type.__args__[1])
    else:
        python_cls = get_python_type(type(thrift_object))
        hints = typing.get_type_hints(python_cls)
        obj = python_cls()
        for thrift_field in dir(thrift_object):
            python_field = camel_to_snake(thrift_field)
            if python_field not in hints:
                continue
            field_type = hints[python_field]
            value = getattr(thrift_object, thrift_field)
            python_value = copy_thrift_to_py(value, field_type)
            if python_value is not None:  # retain default value in object in the case of list and dict
                setattr(obj, python_field, python_value)
    return obj


# sometimes workflow_id maps to either workflowId or workflowID
def last_char_upper(s: str):
    return s[:-1] + s[-1:].upper()


def copy_py_to_thrift(python_object, field_type=None):
    if python_object is None:
        thrift_object = None
    elif field_type and field_type in PRIMITIVES:
        thrift_object = python_object
    elif type(python_object) == list:
        thrift_object = []
        for item in python_object:
            thrift_object.append(copy_py_to_thrift(item, field_type=field_type.__args__[0]))
    elif type(python_object) == dict:
        thrift_object = {}
        assert isinstance(python_object, dict)
        assert field_type
        for key, value in python_object.items():
            thrift_object[key] = copy_py_to_thrift(value, field_type=field_type.__args__[1])
    elif field_type and inspect.isclass(field_type) and issubclass(field_type, IntEnum):
        if python_object is None:
            thrift_object = None
        else:
            assert isinstance(python_object, IntEnum)
            thrift_object = python_object.value
    else:
        thrift_cls = get_thrift_type(type(python_object))
        thrift_object = thrift_cls()
        for python_field, field_type in typing.get_type_hints(type(python_object)).items():
            value = getattr(python_object, python_field)
            thrift_field = snake_to_camel(python_field)
            # Special handling for case of inconsistent naming in shared.thrift
            # StartTimeFilter StartTimeFilter
            if python_field == "start_time_filter":
                thrift_field = "StartTimeFilter"
            elif python_field == "history_event_filter_type":
                thrift_field = "HistoryEventFilterType"
            thrift_value = copy_py_to_thrift(value, field_type)
            if hasattr(thrift_object, thrift_field):
                setattr(thrift_object, thrift_field, thrift_value)
            elif hasattr(thrift_object, last_char_upper(thrift_field)):
                setattr(thrift_object, last_char_upper(thrift_field), thrift_value)
    return thrift_object


def snake_to_camel(snake_str):
    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + ''.join(x.title() for x in components[1:])


def get_thrift_type(python_cls: type) -> type:
    thrift_cls = getattr(cadence_thrift.shared, python_cls.__name__, None)
    assert thrift_cls, "Thrift class not found: " + python_cls.__name__
    return thrift_cls


def get_python_type(thrift_class: type) -> type:
    python_cls = getattr(cadence.cadence_types, thrift_class.__name__, None)
    assert python_cls, "Python class not found: " + thrift_class.__name__
    return python_cls


def args_to_json(args: list) -> str:
    if args is None or len(args) == 0:
        return json.dumps(None)
    elif len(args) == 1:
        return json.dumps(args[0])
    else:
        return json.dumps(args)


def json_to_args(jsonb: bytes) -> typing.Optional[list]:
    parsed = json.loads(jsonb)
    if parsed is None:
        return []
    elif isinstance(parsed, list):
        return parsed
    else:
        return [parsed]



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


@dataclass
class BadRequestError:
    message: str


@dataclass
class InternalServiceError:
    message: str


@dataclass
class DomainAlreadyExistsError:
    message: str


@dataclass
class WorkflowExecutionAlreadyStartedError:
    message: Optional[str]
    startRequestId: Optional[str]
    runId: Optional[str]

    @property
    def start_request_id(self):
        return self.startRequestId

    @property
    def run_id(self):
        return self.runId


@dataclass
class EntityNotExistsError:
    message: str


@dataclass
class ServiceBusyError:
    message: str


@dataclass
class CancellationAlreadyRequestedError:
    message: str


@dataclass
class QueryFailedError:
    message: str


@dataclass
class DomainNotActiveError:
    message: str
    domainName: str
    currentCluster: str
    activeCluster: str

    @property
    def domain_name(self):
        return self.domainName

    @property
    def current_cluster(self):
        return self.currentCluster

    @property
    def active_cluster(self):
        return self.activeCluster


@dataclass
class LimitExceededError:
    message: str


@dataclass
class AccessDeniedError:
    message: str


@dataclass
class RetryTaskError:
    message: str
    domain_id: str
    workflow_id: str
    run_id: str
    next_event_id: int


@dataclass
class ClientVersionNotSupportedError:
    feature_version: str
    client_impl: str
    supported_versions: str


CADENCE_ERROR_FIELDS = {
    "badRequestError": BadRequestError,
    "internalServiceError": InternalServiceError,
    "domainExistsError": DomainAlreadyExistsError,
    "sessionAlreadyExistError": WorkflowExecutionAlreadyStartedError,
    "entityNotExistError": EntityNotExistsError,
    "serviceBusyError": ServiceBusyError,
    "cancellationAlreadyRequestedError": CancellationAlreadyRequestedError,
    "queryFailedError": QueryFailedError,
    "domainNotActiveError": DomainNotActiveError,
    "limitExceededError": LimitExceededError,
    "workflowAlreadyStartedError": WorkflowExecutionAlreadyStartedError,
    "clientVersionNotSupportedError": ClientVersionNotSupportedError
}

IGNORE_FIELDS_IN_ERRORS = ("args", "type_spec", "from_primitive", "to_primitive", "with_traceback")


def find_error(response):
    for key, cls in CADENCE_ERROR_FIELDS.items():
        error = getattr(response, key, None)
        if error:
            kwargs = {}
            for field in dir(error):
                if field not in IGNORE_FIELDS_IN_ERRORS and not field.startswith("__"):
                    kwargs[field] = getattr(error, field)
            return cls(**kwargs)
    return None



from __future__ import annotations
from dataclasses import dataclass
from enum import Enum


class DecisionState(Enum):
    CREATED = 1
    DECISION_SENT = 2
    CANCELED_BEFORE_INITIATED = 3
    INITIATED = 4
    STARTED = 5
    CANCELED_AFTER_INITIATED = 6
    CANCELED_AFTER_STARTED = 7
    CANCELLATION_DECISION_SENT = 8
    COMPLETED_AFTER_CANCELLATION_DECISION_SENT = 9
    COMPLETED = 10


class DecisionTarget(Enum):
    ACTIVITY = 1
    CHILD_WORKFLOW = 2
    CANCEL_EXTERNAL_WORKFLOW = 3
    SIGNAL_EXTERNAL_WORKFLOW = 4
    TIMER = 5
    MARKER = 6

    # Probably won't end up using this since the Python version won't have something analagous to
    # CompleteWorkflowStateMachine
    SELF = 7


@dataclass
class DecisionId:
    decision_target: DecisionTarget
    decision_event_id: int

    def __str__(self):
        return f"{self.decision_target}:{self.decision_event_id}"

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other: DecisionId):
        return (self.decision_target == other.decision_target) and (self.decision_event_id == other.decision_event_id)



from __future__ import annotations
import datetime
import inspect
import json
import random
import uuid
from dataclasses import dataclass, field
from typing import Callable, List, Type, Dict, Tuple
from uuid import uuid4

from cadence.cadence_types import WorkflowIdReusePolicy, StartWorkflowExecutionRequest, TaskList, WorkflowType, \
    GetWorkflowExecutionHistoryRequest, WorkflowExecution, HistoryEventFilterType, EventType, HistoryEvent, \
    StartWorkflowExecutionResponse, SignalWorkflowExecutionRequest
from cadence.conversions import args_to_json
from cadence.workflowservice import WorkflowService


class Workflow:

    @staticmethod
    def new_activity_stub(activities_cls):
        from cadence.decision_loop import ITask
        task: ITask = ITask.current()
        assert task
        cls = activities_cls()
        cls._decision_context = task.decider.decision_context
        return cls

    @staticmethod
    async def await_till(c: Callable, timeout_seconds: int = 0) -> bool:
        from cadence.decision_loop import ITask
        task: ITask = ITask.current()
        assert task
        return await task.await_till(c, timeout_seconds)

    @staticmethod
    async def sleep(seconds: int):
        from cadence.decision_loop import ITask
        task: ITask = ITask.current()
        await task.decider.decision_context.schedule_timer(seconds)

    @staticmethod
    def current_time_millis() -> int:
        from cadence.decision_loop import ITask
        task: ITask = ITask.current()
        return task.decider.decision_context.current_time_millis()

    @staticmethod
    def now() -> datetime.datetime:
        from cadence.decision_loop import ITask
        task: ITask = ITask.current()
        now_in_ms = task.decider.decision_context.current_time_millis()
        return datetime.datetime.fromtimestamp(now_in_ms / 1000)

    @staticmethod
    def random_uuid() -> uuid.UUID:
        from cadence.decision_loop import ITask
        task: ITask = ITask.current()
        return task.decider.decision_context.random_uuid()

    @staticmethod
    def new_random() -> random.Random:
        from cadence.decision_loop import ITask
        task: ITask = ITask.current()
        return task.decider.decision_context.new_random()

class WorkflowStub:
    pass


@dataclass
class WorkflowClient:
    service: WorkflowService
    domain: domain
    options: WorkflowClientOptions

    @classmethod
    def new_client(cls, host: str = "localhost", port: int = 7933, domain: str = "",
                   options: WorkflowClientOptions = None) -> WorkflowClient:
        service = WorkflowService.create(host, port)
        return cls(service=service, domain=domain, options=options)

    @classmethod
    def start(cls, stub_fn: Callable, *args) -> WorkflowExecution:
        stub = stub_fn.__self__
        assert stub._workflow_client is not None
        assert stub_fn._workflow_method is not None
        return exec_workflow(stub._workflow_client, stub_fn._workflow_method, args, workflow_options=stub._workflow_options, stub_instance=stub)

    def new_workflow_stub(self, cls: Type, workflow_options: WorkflowOptions = None):
        attrs = {}
        attrs["_workflow_client"] = self
        attrs["_workflow_options"] = workflow_options
        for name, fn in inspect.getmembers(cls, inspect.isfunction):
            if hasattr(fn, "_workflow_method"):
                attrs[name] = get_workflow_stub_fn(fn._workflow_method)
            elif hasattr(fn, "_signal_method"):
                attrs[name] = get_signal_stub_fn(fn._signal_method)
        stub_cls = type(cls.__name__, (WorkflowStub,), attrs)
        return stub_cls()

    def wait_for_close(self, execution: WorkflowExecution) -> object:
        while True:
            history_request = create_close_history_event_request(self, execution.workflow_id, execution.run_id)
            history_response, err = self.service.get_workflow_execution_history(history_request)
            if err:
                raise Exception(err)
            if not history_response.history.events:
                continue
            history_event = history_response.history.events[0]
            if history_event.event_type == EventType.WorkflowExecutionCompleted:
                attributes = history_event.workflow_execution_completed_event_attributes
                return json.loads(attributes.result)
            elif history_event.event_type == EventType.WorkflowExecutionFailed:
                attributes = history_event.workflow_execution_failed_event_attributes
                details: Dict = json.loads(attributes.details)
                detail_message = details.get("detailMessage", "")
                raise WorkflowExecutionFailedException(attributes.reason, details=details, detail_message=detail_message)
            elif history_event.event_type == EventType.WorkflowExecutionTimedOut:
                raise WorkflowExecutionTimedOutException()
            elif history_event.event_type == EventType.WorkflowExecutionTerminated:
                attributes = history_event.workflow_execution_terminated_event_attributes
                raise WorkflowExecutionTerminatedException(reason=attributes.reason, details=attributes.details,
                                                           identity=attributes.identity)
            elif history_event.event_type == EventType.WorkflowExecutionCanceled:
                raise WorkflowExecutionCanceledException()
            else:
                raise Exception("Unexpected history close event: " + str(history_event))


def exec_workflow(workflow_client, wm: WorkflowMethod, args, workflow_options: WorkflowOptions = None, stub_instance: object = None) -> WorkflowExecution:
    start_request = create_start_workflow_request(workflow_client, wm, args)
    start_response, err = workflow_client.service.start_workflow(start_request)
    if err:
        raise Exception(err)
    execution = WorkflowExecution(workflow_id=start_request.workflow_id, run_id=start_response.run_id)
    stub_instance._execution = execution
    return execution


def exec_workflow_sync(workflow_client: WorkflowClient, wm: WorkflowMethod, args: List,
                       workflow_options: WorkflowOptions = None, stub_instance: object = None):
    execution: WorkflowExecution = exec_workflow(workflow_client, wm, args, workflow_options=workflow_options, stub_instance=stub_instance)
    return workflow_client.wait_for_close(execution)


def exec_signal(workflow_client: WorkflowClient, sm: SignalMethod, args, stub_instance: object = None):
    assert stub_instance._execution
    request = SignalWorkflowExecutionRequest()
    request.workflow_execution = stub_instance._execution
    request.signal_name = sm.name
    request.input = args_to_json(args)
    request.domain = workflow_client.domain
    response, err = workflow_client.service.signal_workflow_execution(request)
    if err:
        raise Exception(err)


def create_start_workflow_request(workflow_client: WorkflowClient, wm: WorkflowMethod,
                                  args: List) -> StartWorkflowExecutionRequest:
    start_request = StartWorkflowExecutionRequest()
    start_request.domain = workflow_client.domain
    start_request.workflow_id = wm._workflow_id if wm._workflow_id else str(uuid4())
    start_request.workflow_type = WorkflowType()
    start_request.workflow_type.name = wm._name
    start_request.task_list = TaskList()
    start_request.task_list.name = wm._task_list
    start_request.input = args_to_json(args)
    start_request.execution_start_to_close_timeout_seconds = wm._execution_start_to_close_timeout_seconds
    start_request.task_start_to_close_timeout_seconds = wm._task_start_to_close_timeout_seconds
    start_request.identity = workflow_client.service.get_identity()
    start_request.workflow_id_reuse_policy = wm._workflow_id_reuse_policy
    start_request.request_id = str(uuid4())
    start_request.cron_schedule = wm._cron_schedule if wm._cron_schedule else None
    return start_request


def create_close_history_event_request(workflow_client: WorkflowClient, workflow_id: str,
                                       run_id: str) -> GetWorkflowExecutionHistoryRequest:
    history_request = GetWorkflowExecutionHistoryRequest()
    history_request.domain = workflow_client.domain
    history_request.execution = WorkflowExecution()
    history_request.execution.workflow_id = workflow_id
    history_request.execution.run_id = run_id
    history_request.wait_for_new_event = True
    history_request.history_event_filter_type = HistoryEventFilterType.CLOSE_EVENT
    return history_request


def get_workflow_method_name(method):
    return "::".join(method.__qualname__.split(".")[-2:])


def get_workflow_stub_fn(wm: WorkflowMethod):
    def workflow_stub_fn(self, *args):
        assert self._workflow_client is not None
        return exec_workflow_sync(self._workflow_client, wm, args,
                                  workflow_options=self._workflow_options, stub_instance=self)
    workflow_stub_fn._workflow_method = wm
    return workflow_stub_fn


def get_signal_stub_fn(sm: SignalMethod):
    def signal_stub_fn(self, *args):
        assert self._workflow_client is not None
        return exec_signal(self._workflow_client, sm, args, stub_instance=self)
    signal_stub_fn._signal_method = sm
    return signal_stub_fn

@dataclass
class WorkflowMethod(object):
    _name: str = None
    _workflow_id: str = None
    _workflow_id_reuse_policy: WorkflowIdReusePolicy = None
    _execution_start_to_close_timeout_seconds: int = None
    _task_start_to_close_timeout_seconds: int = None
    _task_list: str = None
    _cron_schedule: str = None


def workflow_method(func=None,
                    name=None,
                    workflow_id=None,
                    workflow_id_reuse_policy=WorkflowIdReusePolicy.AllowDuplicateFailedOnly,
                    execution_start_to_close_timeout_seconds=7200,  # (2 hours)
                    task_start_to_close_timeout_seconds=10,  # same timeout as Java library
                    task_list=None):
    def wrapper(fn):
        if not hasattr(fn, "_workflow_method"):
            fn._workflow_method = WorkflowMethod()
        fn._workflow_method._name = name if name else get_workflow_method_name(fn)
        fn._workflow_method._workflow_id = workflow_id
        fn._workflow_method._workflow_id_reuse_policy = workflow_id_reuse_policy
        fn._workflow_method._execution_start_to_close_timeout_seconds = execution_start_to_close_timeout_seconds
        fn._workflow_method._task_start_to_close_timeout_seconds = task_start_to_close_timeout_seconds
        fn._workflow_method._task_list = task_list
        return fn
    if func and inspect.isfunction(func):
        return wrapper(func)
    else:
        return wrapper


@dataclass
class SignalMethod:
    name: str = None


def signal_method(func=None, name: str = None):
    def wrapper(fn):
        fn._signal_method = SignalMethod()
        fn._signal_method.name = name if name else get_workflow_method_name(fn)
        return fn
    if func and inspect.isfunction(func):
        return wrapper(func)
    else:
        return wrapper


def cron_schedule(value):
    def wrapper(fn):
        if not hasattr(fn, "_workflow_method"):
            fn._workflow_method = WorkflowMethod()
        fn._workflow_method._cron_schedule = value
        return fn
    return wrapper


@dataclass
class WorkflowClientOptions:
    pass


@dataclass
class WorkflowOptions:
    pass


@dataclass
class WorkflowExecutionFailedException(Exception):
    reason: str
    details: Dict[str, str]
    detail_message: str

    def __str__(self) -> str:
        cause = self.details.get("cause")
        if cause:
            return f"{cause['class']}: {cause['detailMessage']}"
        else:
            return f"{self.reason}: {self.detail_message}"


@dataclass
class WorkflowExecutionTimedOutException(Exception):
    pass


@dataclass
class WorkflowExecutionCanceledException(Exception):
    pass


@dataclass
class WorkflowExecutionTerminatedException(Exception):
    reason: str
    details: object
    identity: str

    def __str__(self) -> str:
        return self.reason



from dataclasses import dataclass, field
from typing import Callable, List, Optional

from cadence.cadence_types import Decision, HistoryEvent, ScheduleActivityTaskDecisionAttributes, \
    RequestCancelActivityTaskDecisionAttributes, DecisionType, StartTimerDecisionAttributes, \
    CancelTimerDecisionAttributes
from cadence.decisions import DecisionState, DecisionId
from cadence.exceptions import IllegalStateException, IllegalArgumentException


class DecisionStateMachine:
    def get_decision(self) -> Optional[Decision]:
        raise NotImplementedError

    def cancel(self, immediate_cancellation_callback: Callable) -> bool:
        raise NotImplementedError

    def handle_started_event(self, event: HistoryEvent):
        raise NotImplementedError

    def handle_cancellation_initiated_event(self):
        raise NotImplementedError

    def handle_cancellation_event(self):
        raise NotImplementedError

    def handle_cancellation_failure_event(self, event: HistoryEvent):
        raise NotImplementedError

    def handle_completion_event(self):
        raise NotImplementedError

    def handle_initiation_failed_event(self, event: HistoryEvent):
        raise NotImplementedError

    def handle_initiated_event(self, event: HistoryEvent):
        raise NotImplementedError

    def handle_decision_task_started_event(self):
        raise NotImplementedError

    def get_state(self) -> DecisionState:
        raise NotImplementedError

    def is_done(self) -> bool:
        raise NotImplementedError

    def get_id(self) -> DecisionId:
        raise NotImplementedError


# noinspection PyAbstractClass
@dataclass
class DecisionStateMachineBase(DecisionStateMachine):
    """
    This class has feature parity with the Java version even though it implements parts of features
    not yet implemented in the Python version.
    """
    id: DecisionId = None
    state: DecisionState = DecisionState.CREATED
    state_history: List[str] = field(default_factory=list)

    def __post_init__(self):
        self.state_history.append(str(self))

    def get_state(self) -> DecisionState:
        return self.state

    def get_id(self) -> DecisionId:
        return self.id

    def is_done(self) -> bool:
        return self.state in (DecisionState.COMPLETED,
                              DecisionState.COMPLETED_AFTER_CANCELLATION_DECISION_SENT)

    def handle_decision_task_started_event(self):
        if self.state == DecisionState.CREATED:
            self.state_history.append("handle_decision_task_started_event")
            self.state = DecisionState.DECISION_SENT
            self.state_history.append(str(self.state))
        else:
            pass

    def cancel(self, immediate_cancellation_callback: Optional[Callable]) -> bool:
        self.state_history.append("cancel")
        result = False
        if self.state == DecisionState.CREATED:
            self.state = DecisionState.COMPLETED
            if immediate_cancellation_callback:
                immediate_cancellation_callback()
        elif self.state == DecisionState.DECISION_SENT:
            self.state = DecisionState.CANCELED_BEFORE_INITIATED
            result = True
        elif self.state == DecisionState.INITIATED:
            self.state = DecisionState.CANCELED_AFTER_INITIATED
            result = True
        else:
            self.fail_state_transition()
        self.state_history.append(str(self.state))
        return result

    def handle_initiated_event(self, event: HistoryEvent):
        self.state_history.append("handle_initiated_event")
        if self.state == DecisionState.DECISION_SENT:
            self.state = DecisionState.INITIATED
        elif self.state == DecisionState.CANCELED_BEFORE_INITIATED:
            self.state = DecisionState.CANCELED_AFTER_INITIATED
        else:
            self.fail_state_transition()
        self.state_history.append(str(self.state))

    def handle_initiation_failed_event(self, event: HistoryEvent):
        self.state_history.append("handle_initiation_failed_event")
        if self.state in (
                DecisionState.INITIATED, DecisionState.DECISION_SENT, DecisionState.CANCELED_BEFORE_INITIATED):
            self.state = DecisionState.COMPLETED
        else:
            self.fail_state_transition()
        self.state_history.append(str(self.state))

    def handle_started_event(self, event: HistoryEvent):
        self.state_history.append("handle_started_event")

    def handle_completion_event(self):
        self.state_history.append("handle_completion_event")
        if self.state in (DecisionState.CANCELED_AFTER_INITIATED, DecisionState.INITIATED):
            self.state = DecisionState.COMPLETED
        elif self.state == DecisionState.CANCELLATION_DECISION_SENT:
            self.state = DecisionState.COMPLETED_AFTER_CANCELLATION_DECISION_SENT
        else:
            self.fail_state_transition()
        self.state_history.append(str(self.state))

    def handle_cancellation_initiated_event(self):
        self.state_history.append("handle_cancellation_initiated_event")
        if self.state == DecisionState.CANCELLATION_DECISION_SENT:
            # No state change
            pass
        else:
            self.fail_state_transition()
        self.state_history.append(str(self.state))

    def handle_cancellation_failure_event(self, event: HistoryEvent):
        self.state_history.append("handle_cancellation_failure_event")
        if self.state == DecisionState.COMPLETED_AFTER_CANCELLATION_DECISION_SENT:
            self.state = DecisionState.COMPLETED
        else:
            self.state.fail_state_transition()
        self.state_history.append(str(self.state))

    def handle_cancellation_event(self):
        self.state_history.append("handle_cancellation_event")
        if self.state == DecisionState.CANCELLATION_DECISION_SENT:
            self.state = DecisionState.COMPLETED
        else:
            self.fail_state_transition()
        self.state_history.append(str(self.state))

    def fail_state_transition(self):
        raise IllegalStateException("id=" + str(self.id) + ", transitions=" + str(self.state_history))


@dataclass
class ActivityDecisionStateMachine(DecisionStateMachineBase):
    """
    This class has feature parity with the Java version even though it implements parts of features
    not yet implemented in the Python version.
    """
    schedule_attributes: ScheduleActivityTaskDecisionAttributes = None

    def __post_init__(self):
        if not self.schedule_attributes:
            raise IllegalArgumentException("schedule_attributes is mandatory")

    def get_decision(self) -> Optional[Decision]:
        if self.state == DecisionState.CREATED:
            return self.create_schedule_activity_task_decision()
        elif self.state == DecisionState.CANCELED_AFTER_INITIATED:
            return self.create_request_cancel_activity_task_decision()
        else:
            return None

    def handle_decision_task_started_event(self):
        if self.state == DecisionState.CANCELED_AFTER_INITIATED:
            self.state_history.append("handle_decision_task_started_event")
            self.state = DecisionState.CANCELLATION_DECISION_SENT
            self.state_history.append(str(self.state))
        else:
            super().handle_decision_task_started_event()

    def handle_cancellation_failure_event(self, event: HistoryEvent):
        if self.state == DecisionState.CANCELLATION_DECISION_SENT:
            self.state_history.append("handle_cancellation_failure_event")
            self.state = DecisionState.INITIATED
            self.state_history.append(str(self.state))
        else:
            super().handle_cancellation_failure_event(event)

    def create_schedule_activity_task_decision(self):
        decision = Decision()
        decision.schedule_activity_task_decision_attributes = self.schedule_attributes
        decision.decision_type = DecisionType.ScheduleActivityTask
        return decision

    def create_request_cancel_activity_task_decision(self):
        try_cancel = RequestCancelActivityTaskDecisionAttributes()
        try_cancel.activity_id = self.schedule_attributes.activity_id
        decision = Decision()
        decision.request_cancel_activity_task_decision_attributes = try_cancel
        decision.decision_type = DecisionType.RequestCancelActivityTask
        return decision


# noinspection PyAbstractClass
@dataclass
class CompleteWorkflowStateMachine(DecisionStateMachine):
    id: DecisionId
    decision: Optional[Decision]

    def get_id(self) -> DecisionId:
        return self.id

    def get_decision(self) -> Optional[Decision]:
        return self.decision

    def handle_initiation_failed_event(self, event: HistoryEvent):
        self.decision = None

    def get_state(self) -> DecisionState:
        return DecisionState.CREATED

    def is_done(self) -> bool:
        return self.decision is not None

    def handle_decision_task_started_event(self):
        pass


# noinspection PyAbstractClass
@dataclass
class TimerDecisionStateMachine(DecisionStateMachineBase):
    start_timer_attributes: StartTimerDecisionAttributes = None
    canceled: bool = False

    def __post_init__(self):
        if not self.start_timer_attributes:
            raise IllegalArgumentException("start_timer_decision_attributes is mandatory")

    def get_decision(self) -> Optional[Decision]:
        if self.state == DecisionState.CREATED:
            return self.create_start_timer_decision()
        elif self.state == DecisionState.CANCELED_AFTER_INITIATED:
            return self.create_cancel_timer_decision()
        else:
            return None

    def handle_decision_task_started_event(self):
        if self.state == DecisionState.CANCELED_AFTER_INITIATED:
            self.state_history.append("handle_decision_task_started_event")
            self.state = DecisionState.CANCELLATION_DECISION_SENT
            self.state_history.append(str(self.state))
        else:
            super().handle_decision_task_started_event()

    def handle_cancellation_failure_event(self, event: HistoryEvent):
        if self.state == DecisionState.CANCELLATION_DECISION_SENT:
            self.state_history.append("handle_cancellation_failure_event")
            self.state = DecisionState.INITIATED
            self.state_history.append(str(self.state))
        else:
            super().handle_cancellation_failure_event(event)

    def cancel(self, immediate_cancellation_callback: Optional[Callable]) -> bool:
        self.canceled = True
        immediate_cancellation_callback()
        return super().cancel(None)

    def is_done(self) -> bool:
        return self.state == DecisionState.COMPLETED or self.canceled

    def create_cancel_timer_decision(self):
        try_cancel = CancelTimerDecisionAttributes()
        try_cancel.timer_id = self.start_timer_attributes.timer_id
        decision: Decision = Decision()
        decision.cancel_timer_decision_attributes = try_cancel
        decision.decision_type = DecisionType.CancelTimer
        return decision

    def create_start_timer_decision(self):
        decision: Decision = Decision()
        decision.start_timer_decision_attributes = self.start_timer_attributes
        decision.decision_type = DecisionType.StartTimer
        return decision



import copy
import inspect
import json
from dataclasses import dataclass
from typing import Callable, List

from cadence.cadence_types import ActivityType
from cadence.conversions import args_to_json


def get_activity_method_name(method: Callable):
    return "::".join(method.__qualname__.split(".")[-2:])


@dataclass
class RetryParameters:
    initial_interval_in_seconds: int
    backoff_coefficient: float
    maximum_interval_in_seconds: int
    maximum_attempts: int
    non_retriable_error_reasons: List[str]
    expiration_interval_in_seconds: int


@dataclass
class ExecuteActivityParameters:
    activity_id: str = ""
    activity_type: ActivityType = None
    heartbeat_timeout_seconds: int = 0
    input: bytes = None
    schedule_to_close_timeout_seconds: int = 0
    schedule_to_start_timeout_seconds: int = 0
    start_to_close_timeout_seconds: int = 0
    task_list: str = ""
    retry_parameters: RetryParameters = None


def activity_method(func: Callable = None, name: str = "", schedule_to_close_timeout_seconds: int = 0,
                    schedule_to_start_timeout_seconds: int = 0, start_to_close_timeout_seconds: int = 0,
                    heartbeat_timeout_seconds: int = 0, task_list: str = ""):
    def wrapper(fn: Callable):
        # noinspection PyProtectedMember
        async def stub_activity_fn(self, *args):
            assert self._decision_context
            assert stub_activity_fn._execute_parameters
            parameters = copy.deepcopy(stub_activity_fn._execute_parameters)
            parameters.input = args_to_json(args).encode("utf-8")
            from cadence.decision_loop import DecisionContext
            decision_context: DecisionContext = self._decision_context
            return await decision_context.schedule_activity_task(parameters=parameters)

        if not task_list:
            raise Exception("task_list parameter is mandatory")

        execute_parameters = ExecuteActivityParameters()
        execute_parameters.activity_type = ActivityType()
        execute_parameters.activity_type.name = name if name else get_activity_method_name(fn)
        execute_parameters.schedule_to_close_timeout_seconds = schedule_to_close_timeout_seconds
        execute_parameters.schedule_to_start_timeout_seconds = schedule_to_start_timeout_seconds
        execute_parameters.start_to_close_timeout_seconds = start_to_close_timeout_seconds
        execute_parameters.heartbeat_timeout_seconds = heartbeat_timeout_seconds
        execute_parameters.task_list = task_list
        stub_activity_fn._execute_parameters = execute_parameters
        return stub_activity_fn

    if func and inspect.isfunction(func):
        raise Exception("activity_method must be called with arguments")
    else:
        return wrapper



import datetime
import logging
import json

from cadence.cadence_types import PollForActivityTaskRequest, TaskListMetadata, TaskList, PollForActivityTaskResponse, \
    RespondActivityTaskCompletedRequest, RespondActivityTaskFailedRequest
from cadence.conversions import json_to_args
from cadence.workflowservice import WorkflowService

logger = logging.getLogger(__name__)


def activity_task_loop(worker):
    service = WorkflowService.create(worker.host, worker.port)
    logger.info(f"Activity task worker started: {WorkflowService.get_identity()}")
    try:
        while True:
            if worker.is_stop_requested():
                return
            try:
                polling_start = datetime.datetime.now()
                polling_request = PollForActivityTaskRequest()
                polling_request.task_list_metadata = TaskListMetadata()
                polling_request.task_list_metadata.max_tasks_per_second = 200000
                polling_request.domain = worker.domain
                polling_request.identity = WorkflowService.get_identity()
                polling_request.task_list = TaskList()
                polling_request.task_list.name = worker.task_list
                task: PollForActivityTaskResponse
                task, err = service.poll_for_activity_task(polling_request)
                polling_end = datetime.datetime.now()
                logger.debug("PollForActivityTask: %dms", (polling_end - polling_start).total_seconds() * 1000)
            except Exception as ex:
                logger.error("PollForActivityTask error: %s", ex)
                continue
            if err:
                logger.error("PollForActivityTask failed: %s", err)
                continue
            if not task.task_token:
                logger.debug("PollForActivityTask has no task_token (expected): %s", task)
                continue

            args = json_to_args(task.input)
            logger.info(f"Request for activity: {task.activity_type.name}")
            fn = worker.activities.get(task.activity_type.name)
            if not fn:
                logger.error("Activity type not found: " + task.activity_type.name)
                continue

            process_start = datetime.datetime.now()
            try:
                ret = fn(*args)
                respond = RespondActivityTaskCompletedRequest()
                respond.task_token = task.task_token
                respond.result = json.dumps(ret)
                respond.identity = WorkflowService.get_identity()
                _, error = service.respond_activity_task_completed(respond)
                if error:
                    logger.error("Error invoking RespondActivityTaskCompleted: %s", error)
                logger.info(f"Activity {task.activity_type.name}({str(args)[1:-1]}) returned {respond.result}")
            except Exception as ex:
                logger.error(f"Activity {task.activity_type.name} failed: {type(ex).__name__}({ex})", exc_info=1)
                respond: RespondActivityTaskFailedRequest = RespondActivityTaskFailedRequest()
                respond.task_token = task.task_token
                respond.identity = WorkflowService.get_identity()
                respond.details = json.dumps({
                    "detailMessage": f"Python error: {type(ex).__name__}({ex})",
                    "class": "java.lang.Exception"
                })
                respond.reason = "java.lang.Exception"
                _, error = service.respond_activity_task_failed(respond)
                if error:
                    logger.error("Error invoking RespondActivityTaskFailed: %s", error)

            process_end = datetime.datetime.now()
            logger.info("Process ActivityTask: %dms", (process_end - process_start).total_seconds() * 1000)
    finally:
        worker.notify_thread_stopped()



from __future__ import annotations
from .ioutils import IOWrapper


class KVHeaders:
    @classmethod
    def read_kv_headers(cls, fp: IOWrapper, len_size, field: str) -> KVHeaders:
        if len_size == 2:
            fn = fp.read_short
        elif len_size == 1:
            fn = fp.read_byte
        nh = fn(field + "." + "nh")
        d = {}
        for i in range(nh):
            key_len = fn("field[%d].key_len" % (i + 1))
            key = fp.read_string(key_len, "field[%d].key" % (i + 1))
            value_len = fn("field[%d].value_len" % (i + 1))
            value = fp.read_string(value_len, "field[%d].value" % (i + 1))
            d[key] = value
        return KVHeaders(d, len_size)

    def __init__(self, d=None, len_size=2):
        if not d:
            d = {}
        self.d = d
        self.len_size = len_size

    def write_headers(self, fp: IOWrapper):
        if self.len_size == 2:
            fn = fp.write_short
        elif self.len_size == 1:
            fn = fp.write_byte
        fn(len(self.d))
        for key, value in self.d.items():
            fn(len(key))
            fp.write_string(key)
            fn(len(value))
            fp.write_string(value)

    def size(self):
        return self.len_size + (self.len_size * len(self.d) * 2) + sum([len(k) for k in self.d.keys()]) + sum(
            [len(v) for v in self.d.values()])



from typing import IO


class IOWrapper:
    io_stream: IO

    def __init__(self, io_stream: IO):
        self.io_stream = io_stream

    def read_or_eof(self, size, field):
        buf: bytes = self.io_stream.read(size)
        if len(buf) != size:
            raise EOFError(field)
        return buf

    def read_short(self, field: str) -> int:
        return int.from_bytes(self.read_or_eof(2, field), byteorder='big', signed=False)

    def read_long(self, field: str) -> int:
        return int.from_bytes(self.read_or_eof(4, field), byteorder='big', signed=False)

    def read_byte(self, field: str) -> int:
        return int.from_bytes(self.read_or_eof(1, field), byteorder='big', signed=False)

    def read_bytes(self, n: int, field: str) -> bytes:
        buf: bytes = self.read_or_eof(n, field)
        return buf

    def read_string(self, n: int, field: str) -> str:
        buf: bytes = self.read_or_eof(n, field)
        return str(buf, "utf-8")

    def write_short(self, v: int):
        self.io_stream.write(v.to_bytes(2, byteorder='big', signed= False))

    def write_long(self, v: int):
        self.io_stream.write(v.to_bytes(4, byteorder='big', signed=False))

    def write_byte(self, v: int):
        self.io_stream.write(v.to_bytes(1, byteorder='big', signed=False))

    def write_bytes(self, b: bytes):
        self.io_stream.write(b)

    def write_string(self, s: str):
        self.io_stream.write(bytes(s, "utf-8"))

    def flush(self):
        self.io_stream.flush()



from dataclasses import dataclass
from typing import Callable


@dataclass
class OpenRequestInfo:
    # BiConsumer<T, Exception>
    completion_handle: Callable = None
    user_context: object = None




from __future__ import annotations

from typing import Type, Dict, Optional, IO, List

from .ioutils import IOWrapper
from .kvheaders import KVHeaders

FRAME_HEADER_SIZE = 16


# Helper class to make get_payload_size() functions more readable
class LenHelper(object):
    def __getattr__(self, name):
        def fn(*args):
            return args[0]
        return fn


_ = LenHelper()


class FrameHeader:
    """
    size:2 type:1 reserved:1 id:4
    reserved:8
    """
    @classmethod
    def read_header(cls, fp: IOWrapper) -> FrameHeader:
        size = fp.read_short("header.size")
        payload_type = fp.read_byte("header.payload_type")
        reserved1 = fp.read_byte("header.reserved1")
        frame_id = fp.read_long("header.id")
        reserved2 = fp.read_bytes(8, "header.reserved2")
        return cls(size, payload_type, frame_id)

    def __init__(self, size: int = None, payload_type: int = None, frame_id: int = None):
        self.size = size
        self.payload_type = payload_type
        self.id = frame_id


class Frame:
    TYPE = None

    @staticmethod
    def read_frame(wrapper: IOWrapper) -> Frame:
        header: FrameHeader = FrameHeader.read_header(wrapper)
        frame_types: Dict[int, Type[Frame]] = {
            0x01: InitReqFrame,
            0x02: InitResFrame,
            0x03: CallReqFrame,
            0x04: CallResFrame,
            0x13: CallReqContinueFrame,
            0x14: CallResContinueFrame,
            0Xff: ErrorFrame
        }

        frame_cls: Optional[Type[Frame]] = frame_types.get(header.payload_type)
        if not frame_cls:
            raise NotImplementedError("Payload type: %s " % hex(header.payload_type))
        frame: Frame = frame_cls()
        frame.read(header, wrapper)
        return frame

    def __init__(self):
        self.id = None

    def read(self, header: FrameHeader, fp: IOWrapper):
        self.id = header.id
        self.read_payload(fp, header.size - FRAME_HEADER_SIZE)

    def read_payload(self, fp: IOWrapper, size: int):
        raise NotImplementedError()

    def write(self, wrapper: IOWrapper):
        self.write_header(wrapper)
        self.write_payload(wrapper)

    def write_header(self, fp: IOWrapper):
        fp.write_short(self.get_size())
        fp.write_byte(self.TYPE)
        fp.write_byte(0)
        fp.write_long(self.id)
        fp.write_bytes(bytes(8))

    def get_size(self):
        return FRAME_HEADER_SIZE + self.get_payload_size()

    def get_payload_size(self):
        raise NotImplementedError()

    def write_payload(self, fp: IOWrapper):
        raise NotImplementedError()

    def is_full(self):
        size = self.get_size()
        assert size <= 0xFFFF
        return size == 0xFFFF

    def space_available(self):
        size = self.get_size()
        return 0xFFFF - size

    def has_space_available(self, n):
        return n >= (0xFFFF - self.get_size())

    def is_frame_boundary(self):
        # Frame cannot fit any more arguments
        return self.has_space_available(3)


class InitReqFrame(Frame):
    VERSION = 2
    TYPE = 0x01

    def __init__(self):
        super().__init__()
        self.headers = KVHeaders(len_size=2)
        self.version = self.VERSION

    def read_payload(self, fp: IOWrapper, size: int):
        self.version = fp.read_short("initreq.version")
        self.headers = KVHeaders.read_kv_headers(fp, 2, "initreq.headers")

    def get_payload_size(self):
        return 2 + self.headers.size()

    def write_payload(self, fp: IOWrapper):
        fp.write_short(self.VERSION)
        self.headers.write_headers(fp)


class InitResFrame(Frame):
    VERSION = 2
    TYPE = 0x02

    def __init__(self):
        super().__init__()
        self.headers = KVHeaders(len_size=2)
        self.version = self.VERSION

    def read_payload(self, fp: IOWrapper, size: int):
        self.version = fp.read_short("initres.version")
        self.headers = KVHeaders.read_kv_headers(fp, 2, "initres.headers")

    def get_payload_size(self):
        return 2 + self.headers.size()

    def write_payload(self, fp: IOWrapper):
        fp.write_short(self.VERSION)
        self.headers.write_headers(fp)


class Arg:

    @classmethod
    def read_arg(cls, fp: IOWrapper, offset, payload_size, possible_fragment, field):
        arg_length = fp.read_short(field + ".arg_length")
        offset += 2
        buf = fp.read_bytes(arg_length, field + ".arg")
        offset += arg_length
        is_fragment = False
        # if there is more data
        if possible_fragment and offset < payload_size:
            is_fragment = False
        elif possible_fragment and offset == payload_size:
            is_fragment = True
        elif possible_fragment:
            raise Exception("Malformed frame argument")
        return cls(buf, is_fragment)

    def __init__(self, buf, is_fragment=False):
        self.buf = buf
        self.is_fragment = is_fragment

    def size(self):
        return 2 + len(self.buf)

    def write_arg(self, fp: IOWrapper):
        fp.write_short(len(self.buf))
        fp.write_bytes(self.buf)


class CallFlags:
    FLAG_MORE_FRAGMENTS_FOLLOW = 0x01
    # noinspection PyPep8
    FLAG_REQUEST_STREAMING     = 0x02

    def __init__(self):
        self.flags = 0

    def set_more_fragments_follow(self, b):
        if b:
            self.flags = self.flags | self.FLAG_MORE_FRAGMENTS_FOLLOW
        else:
            self.flags = self.flags & ~self.FLAG_MORE_FRAGMENTS_FOLLOW

    def is_more_fragments_follow(self):
        return self.flags & self.FLAG_MORE_FRAGMENTS_FOLLOW

    def set_request_streaming(self, b):
        if b:
            self.flags = self.flags | self.FLAG_REQUEST_STREAMING
        else:
            self.flags = self.flags & ~self.FLAG_REQUEST_STREAMING

    def is_request_streaming(self):
        return self.flags & self.FLAG_REQUEST_STREAMING


class FrameWithArgs(Frame):
    args: List[Arg]

    def __init__(self):
        self.args = []


class CallReqFrame(FrameWithArgs, CallFlags):
    """
    flags:1 ttl:4 tracing:25
    service~1 nh:1 (hk~1 hv~1){nh}
    csumtype:1 (csum:4){0,1} arg1~2 arg2~2 arg3~2
    """
    TYPE = 0x03

    def __init__(self):
        FrameWithArgs.__init__(self)
        CallFlags.__init__(self)
        self.ttl = 0
        self.tracing = bytes(25)
        self.service = ""
        self.headers = KVHeaders({}, 1)
        self.csumtype = 0

    # noinspection PyPep8
    def read_payload(self, fp: IOWrapper, size: int):
        offset = 0
        self.flags = fp.read_byte("callreq.flags"); offset += 1;
        self.ttl = fp.read_long("callreq.ttl"); offset += 4;
        self.tracing = fp.read_bytes(25, "callreq.tracing"); offset += 25;
        service_len = fp.read_byte('callreq.service_len'); offset += 1;
        self.service = fp.read_string(service_len, "callreq.service"); offset += service_len;
        self.headers = KVHeaders.read_kv_headers(fp, 1, "callreq.headers"); offset += self.headers.size();
        self.csumtype = fp.read_byte("callreq.csumtype"); offset += 1;
        if self.csumtype != 0:
            raise NotImplementedError("Checksum type not supported")
        self.args = []
        arg_count = 1
        while offset < size:
            arg = Arg.read_arg(fp, offset, size, self.is_more_fragments_follow(), "callreq.args[%d]" % arg_count)
            self.args.append(arg); offset += arg.size();
            arg_count += 1

    def get_payload_size(self):
        return (_.flags(1) + _.ttl(4) + _.tracing(25) +
                _.service_len(1) + len(self.service) + self.headers.size() +
                _.csumtype(1) + sum([arg.size() for arg in self.args]))

    # noinspection PyTrailingSemicolon,PyPep8
    def write_payload(self, fp: IOWrapper):
        offset = 0
        fp.write_byte(self.flags); offset += 1;
        fp.write_long(self.ttl); offset += 4;
        fp.write_bytes(self.tracing); offset += 25;
        fp.write_byte(len(self.service)); offset += 1;
        fp.write_string(self.service); offset += len(self.service);
        self.headers.write_headers(fp); offset += self.headers.size();
        fp.write_byte(self.csumtype); offset += 1;
        for arg in self.args:
            arg.write_arg(fp); offset += arg.size();
        assert offset == self.get_payload_size()


class CallResFrame(FrameWithArgs, CallFlags):
    """
    flags:1 code:1 tracing:25
    nh:1 (hk~1 hv~1){nh}
    csumtype:1 (csum:4){0,1} arg1~2 arg2~2 arg3~2
    """
    TYPE = 0x04

    def __init__(self):
        FrameWithArgs.__init__(self)
        CallFlags.__init__(self)
        self.code = 0
        self.tracing = bytes(25)
        self.headers = KVHeaders({}, 1)
        self.csumtype = 0

    # noinspection PyPep8
    def read_payload(self, fp: IOWrapper, size: int):
        offset = 0
        self.flags = fp.read_byte("callres.flags"); offset += 1;
        self.code = fp.read_byte("callres.code"); offset += 1;
        self.tracing = fp.read_bytes(25, "callres.tracing"); offset += 25;
        self.headers = KVHeaders.read_kv_headers(fp, 1, "callreq.headers"); offset += self.headers.size();
        self.csumtype = fp.read_byte("callres.csumtype"); offset += 1;
        if self.csumtype != 0:
            raise NotImplementedError("Checksum type not supported")
        self.args = []
        arg_count = 1
        while offset < size:
            arg = Arg.read_arg(fp, offset, size, self.is_more_fragments_follow(), "callres.args[%d]" % arg_count)
            self.args.append(arg); offset += arg.size();
            arg_count += 1

    def get_payload_size(self):
        return (_.flag(1) + _.code(1) + _.tracing(25) +
                self.headers.size() +
                _.csumtype(1) + sum([arg.size() for arg in self.args]))

    # noinspection PyPep8
    def write_payload(self, fp: IOWrapper):
        offset = 0
        fp.write_byte(self.flags); offset += 1;
        fp.write_byte(self.code); offset += 1;
        fp.write_bytes(self.tracing); offset += 25;
        self.headers.write_headers(fp); offset += self.headers.size();
        fp.write_byte(self.csumtype); offset += 1;
        for arg in self.args:
            arg.write_arg(fp); offset += arg.size();
        assert offset == self.get_payload_size()


class FrameContinue(FrameWithArgs, CallFlags):
    """
    flags:1 csumtype:1 (csum:4){0,1} {continuation}
    """
    PREFIX = ""

    def __init__(self):
        FrameWithArgs.__init__(self)
        CallFlags.__init__(self)
        self.csumtype = 0

    # noinspection PyPep8
    def read_payload(self, fp: IOWrapper, size: int):
        offset = 0
        self.flags = fp.read_byte(self.PREFIX + ".flags"); offset += 1;
        self.csumtype = fp.read_byte(self.PREFIX + ".csumtype"); offset += 1;
        if self.csumtype != 0:
            raise NotImplementedError("Checksum type not supported")
        self.args = []
        arg_count = 1
        while offset < size:
            arg = Arg.read_arg(fp, offset, size, self.is_more_fragments_follow(), self.PREFIX + ".args[%d]" % arg_count)
            self.args.append(arg); offset += arg.size();
            arg_count += 1

    def get_payload_size(self):
        return _.flag(1) + _.csumtype(1) + sum([arg.size() for arg in self.args])

    # noinspection PyPep8
    def write_payload(self, fp: IOWrapper):
        offset = 0
        fp.write_byte(self.flags); offset += 1;
        fp.write_byte(self.csumtype); offset += 1;
        for arg in self.args:
            arg.write_arg(fp); offset += arg.size();
        assert offset == self.get_payload_size()


class CallReqContinueFrame(FrameContinue):
    PREFIX = "callreqcontinue"
    TYPE = 0x13


class CallResContinueFrame(FrameContinue):
    PREFIX = "callrescontinue"
    TYPE = 0x14


class ErrorFrame(Frame):
    """
    code:1 tracing:25 message~2
    """
    TYPE = 0xff

    message: str
    tracing: bytes
    code: int

    def __init__(self):
        super().__init__()
        self.code = 0
        self.tracing = bytes(25)
        self.message = ""

    def read_payload(self, fp: IOWrapper, size: int):
        offset = 0
        self.code = fp.read_byte("error.code"); offset += 1;
        self.tracing = fp.read_bytes(25, "error.tracing"); offset += 25;
        message_len = fp.read_short("error.message_len"); offset += 2;
        self.message = fp.read_string(message_len, "error.message"); offset += message_len;
        assert offset == size

    def get_payload_size(self):
        return _.code(1) + _.tracing(25) + _.message_len(2) + len(self.message)

    def write_payload(self, fp: IOWrapper):
        offset = 0
        fp.write_byte(self.code); offset += 1;
        fp.write_bytes(self.tracing); offset += 25;
        fp.write_short(len(self.message)); offset += 2;
        fp.write_string(self.message); offset += len(self.message)



from dataclasses import dataclass, field
from typing import List

from cadence.worker import Worker, WorkerOptions


@dataclass
class WorkerFactoryOptions:
    pass


@dataclass
class WorkerFactory:
    host: str = None
    port: int = None
    domain: str = None
    options: WorkerFactoryOptions = None
    workers: List[Worker] = field(default_factory=list)

    def new_worker(self, task_list: str, worker_options: WorkerOptions = None) -> Worker:
        worker = Worker(host=self.host, port=self.port, domain=self.domain, task_list=task_list, options=worker_options)
        self.workers.append(worker)
        return worker

    def start(self):
        for worker in self.workers:
            worker.start()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cadence-client",
    version="0.0.3",
    author="Mohammed Firdaus",
    author_email="firdaus.halim@gmail.com",
    description="Python framework for Cadence Workflow Service",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/firdaus/cadence-python",
    packages=setuptools.find_packages(exclude=["cadence.tests", "cadence.spikes"]),
    install_requires=["more-itertools>=7.0.0", "thriftrw>=1.7.2"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True
)
